#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import math

from qtpy.QtWidgets import QLineEdit, QLabel
from qtpy.QtGui import QIntValidator

from .FramesWidget import FramesWidget


class CorrelationFramesWidget(FramesWidget):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self._last_frame <= 0:
            c_frames = "1"
        else:
            c_frames = str(math.ceil(self._last_frame * 0.5))
        field = QLineEdit(c_frames, self._base)
        validator = QIntValidator(field)

        field.setValidator(validator)
        field.textChanged.connect(self.updateValue)
        field.setPlaceholderText(c_frames)

        self._layout.addWidget(QLabel("Correlation frames", self._base), 1, 0)
        self._layout.addWidget(field, 1, 1)

        self._fields.append(field)
        self._validators.append(QIntValidator(field))
        self._default_values.append(c_frames)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QVBoxLayout,
    QApplication,
    QMenu,
    QLineEdit,
    QTableView,
)
from qtpy.QtCore import Signal, Slot, Qt, QSortFilterProxyModel
from qtpy.QtGui import QStandardItem, QStandardItemModel

from MDANSE.Chemistry import ATOMS_DATABASE
from MDANSE.MLogging import LOG

from MDANSE_GUI.Widgets.GeneralWidgets import InputVariable, InputDialog


class ElementView(QTableView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setSortingEnabled(True)

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        self.clickedPos = event.pos()
        Action1 = menu.addAction("New Atom")
        Action2 = menu.addAction("New Property")
        temp_model = self.model().sourceModel()
        if temp_model is not None:
            Action1.triggered.connect(temp_model.new_line_dialog)
            Action2.triggered.connect(temp_model.new_column_dialog)
        menu.exec_(event.globalPos())


class NewElementDialog(QDialog):
    got_name = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        layout = QVBoxLayout(self)
        edit = QLineEdit(self)
        layout.addWidget(edit)
        self.setLayout(layout)
        self.textedit = edit
        self.button = QPushButton("Accept!", self)
        self.button.clicked.connect(self.accept)
        self.accepted.connect(self.return_value)
        layout.addWidget(self.button)

    @Slot()
    def return_value(self):
        self.got_name.emit(self.textedit.text())


class ElementModel(QStandardItemModel):
    def __init__(self, *args, element_database=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.database = element_database
        self.parseDatabase()

        self.itemChanged.connect(self.write_to_database)

    def parseDatabase(self):
        all_column_names = self.database.properties
        all_row_names = self.database.atoms

        for entry in all_row_names:
            row = []
            atom_info = self.database[entry]
            for key in all_column_names:
                item = QStandardItem(str(atom_info[key]))
                try:
                    intnum = int(str(atom_info[key]))
                except ValueError:
                    try:
                        floatnum = float(atom_info[key])
                    except ValueError:
                        pass
                    except TypeError:
                        pass
                    else:
                        item.setData(floatnum, role=Qt.ItemDataRole.DisplayRole)
                except TypeError:
                    pass
                else:
                    item.setData(intnum, role=Qt.ItemDataRole.DisplayRole)
                row.append(item)
            self.appendRow(row)
        self.setHorizontalHeaderLabels(all_column_names)
        self.setVerticalHeaderLabels(all_row_names)

        self.all_column_names = all_column_names
        self.all_row_names = all_row_names

    @Slot("QStandardItem*")
    def write_to_database(self, item: "QStandardItem"):
        data = item.data()
        text = item.text()
        row = item.row()
        column = item.column()
        LOG.info(f"data:{data}, text:{text},row:{row},column:{column}")
        LOG.info(
            f"column name={self.all_column_names[column]}, row name={self.all_row_names[row]}"
        )
        self.database.set_value(
            self.all_row_names[row], self.all_column_names[column], text
        )
        self.save_changes()

    @Slot()
    def save_changes(self):
        self.database.save()

    @Slot()
    def new_line_dialog(self):
        dialog_variables = [
            InputVariable(
                input_dict={
                    "keyval": "atom_name",
                    "format": str,
                    "label": "New element name",
                    "tooltip": "Type the name of the new chemical element here.",
                    "values": [""],
                }
            )
        ]
        ne_dialog = InputDialog(fields=dialog_variables)
        ne_dialog.got_values.connect(self.add_new_line)
        ne_dialog.show()
        result = ne_dialog.exec()

    @Slot()
    def new_column_dialog(self):
        dialog_variables = [
            InputVariable(
                input_dict={
                    "keyval": "property_name",
                    "format": str,
                    "label": "New property name",
                    "tooltip": "Type the name of the new property here; it will be added to the table.",
                    "values": [""],
                }
            ),
            InputVariable(
                input_dict={
                    "keyval": "property_type",
                    "format": str,
                    "label": "Type of the new property",
                    "tooltip": "One of the following: int, float, str, list",
                    "values": ["float"],
                }
            ),
        ]
        ne_dialog = InputDialog(fields=dialog_variables)
        ne_dialog.got_values.connect(self.add_new_column)
        ne_dialog.show()
        result = ne_dialog.exec()

    @Slot(dict)
    def add_new_line(self, input_variables: dict):
        new_label = "Xx"
        try:
            new_label = input_variables["atom_name"]
        except KeyError:
            return None
        if not new_label in self.database.atoms:
            self.database.add_atom(new_label)
            row = []
            for key in self.all_column_names:
                new_value = self.database.get_value(new_label, key)
                self.database.set_value(new_label, key, new_value)
                item = QStandardItem(str(new_value))
                row.append(item)
            self.all_row_names.append(new_label)
            self.appendRow(row)
            self.setVerticalHeaderItem(
                self.rowCount() - 1, QStandardItem(str(new_label))
            )
            LOG.info(f"self.all_row_names has length: {len(self.all_row_names)}")

    @Slot(dict)
    def add_new_column(self, input_variables: dict):
        new_label = "Xx"
        new_type = "float"
        try:
            new_label = input_variables["property_name"]
        except KeyError:
            return None
        try:
            new_type = input_variables["property_type"]
        except KeyError:
            return None
        if not new_label in self.database.atoms:
            self.database.add_property(new_label, new_type)
            column = []
            for key in self.all_row_names:
                new_value = self.database.get_value(key, new_label)
                self.database.set_value(key, new_label, new_value)
                item = QStandardItem(str(new_value))
                column.append(item)
            self.all_column_names.append(new_label)
            self.appendColumn(column)
            self.setHorizontalHeaderItem(
                self.columnCount() - 1, QStandardItem(str(new_label))
            )
            LOG.info(f"self.all_column_names has length: {len(self.all_column_names)}")


class ElementsDatabaseEditor(QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("MDANSE Chemical Elements Database Editor")

        layout = QVBoxLayout(self)

        self.setLayout(layout)

        self.viewer = ElementView(self)
        layout.addWidget(self.viewer)

        self.data_model = ElementModel(self, element_database=ATOMS_DATABASE)

        self.proxy_model = QSortFilterProxyModel(self)

        self.proxy_model.setSourceModel(self.data_model)
        self.viewer.setModel(self.proxy_model)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    root = ElementsDatabaseEditor()
    root.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from functools import partial

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QWidget

from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.MultiPanel import MultiPanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Views.PlotDetailsView import PlotDetailsView
from MDANSE_GUI.Tabs.Visualisers.PlotHolder import PlotHolder
from MDANSE_GUI.Tabs.Visualisers.PlotSettings import PlotSettings
from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


label_text = """This tab manages a single plot.
You can change the contents and appearance of the plot
here.
"""


class PlotTab(GeneralTab):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._visualiser._unit_lookup = self
        self._visualiser._settings = self._settings
        self._core._extra_visualiser._settings = self._settings
        self._visualiser.currentChanged.connect(self.switch_model)
        self._view.setModel(self.model)
        self._core._extra_visualiser.plot_settings_changed.connect(
            self._visualiser.update_plots
        )
        self._core._extra_visualiser.plot_settings_changed.connect(
            self.model.regenerate_colours
        )
        self._core._extra_visualiser.make_layout()

    @classmethod
    def standard_instance(cls):
        plt_settings = PlotSettings()
        the_tab = cls(
            window,
            name="Plotting",
            session=LocalSession(),
            model=PlottingContext(),
            view=PlotDetailsView(),
            visualiser=PlotHolder(),
            layout=partial(
                MultiPanel, left_panels=[plt_settings], extra_visualiser=plt_settings
            ),
            label_text=label_text,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        plt_settings = PlotSettings(settings=settings)
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            model=None,
            view=PlotDetailsView(),
            visualiser=PlotHolder(),
            layout=partial(
                MultiPanel, left_panels=[plt_settings], extra_visualiser=plt_settings
            ),
            label_text=label_text,
        )
        the_tab._visualiser._unit_lookup = the_tab
        return the_tab

    @property
    def model(self):
        return self._visualiser.model

    @Slot(object)
    def accept_external_data(self, data_model):
        LOG.debug(f"accept_external_data received {data_model}")
        try:
            self._visualiser.model.accept_external_data(data_model)
        except Exception as e:
            LOG.error(f"Visualiser failed to pass data model: {e}")
        else:
            try:
                self._visualiser.plotter.plot_data()
            except Exception as e2:
                LOG.error(f"Visualiser failed to plot data: {e2}")

    @Slot(int)
    def switch_model(self, tab_id):
        self._view.setModel(self.model)


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow

    app = QApplication(sys.argv)
    window = QMainWindow()
    the_tab = PlotTab.standard_instance()
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import (
    QDialog,
    QToolButton,
    QFrame,
    QGridLayout,
    QVBoxLayout,
    QWidget,
    QLabel,
    QApplication,
    QSizePolicy,
    QMenu,
)
from qtpy.QtCore import Signal, Slot, Qt, QPoint
from qtpy.QtGui import QEnterEvent

from MDANSE.Chemistry import ATOMS_DATABASE

# from MDANSE.GUI.ElementsDatabaseEditor import ElementsDatabaseEditor

_LAYOUT = {}
_LAYOUT["H"] = (0, 1)
_LAYOUT["He"] = (0, 18)

_LAYOUT["Li"] = (1, 1)
_LAYOUT["Be"] = (1, 2)
_LAYOUT["B"] = (1, 13)
_LAYOUT["C"] = (1, 14)
_LAYOUT["N"] = (1, 15)
_LAYOUT["O"] = (1, 16)
_LAYOUT["F"] = (1, 17)
_LAYOUT["Ne"] = (1, 18)

_LAYOUT["Na"] = (2, 1)
_LAYOUT["Mg"] = (2, 2)
_LAYOUT["Al"] = (2, 13)
_LAYOUT["Si"] = (2, 14)
_LAYOUT["P"] = (2, 15)
_LAYOUT["S"] = (2, 16)
_LAYOUT["Cl"] = (2, 17)
_LAYOUT["Ar"] = (2, 18)

_LAYOUT["K"] = (3, 1)
_LAYOUT["Ca"] = (3, 2)
_LAYOUT["Sc"] = (3, 3)
_LAYOUT["Ti"] = (3, 4)
_LAYOUT["V"] = (3, 5)
_LAYOUT["Cr"] = (3, 6)
_LAYOUT["Mn"] = (3, 7)
_LAYOUT["Fe"] = (3, 8)
_LAYOUT["Co"] = (3, 9)
_LAYOUT["Ni"] = (3, 10)
_LAYOUT["Cu"] = (3, 11)
_LAYOUT["Zn"] = (3, 12)
_LAYOUT["Ga"] = (3, 13)
_LAYOUT["Ge"] = (3, 14)
_LAYOUT["As"] = (3, 15)
_LAYOUT["Se"] = (3, 16)
_LAYOUT["Br"] = (3, 17)
_LAYOUT["Kr"] = (3, 18)

_LAYOUT["Rb"] = (4, 1)
_LAYOUT["Sr"] = (4, 2)
_LAYOUT["Y"] = (4, 3)
_LAYOUT["Zr"] = (4, 4)
_LAYOUT["Nb"] = (4, 5)
_LAYOUT["Mo"] = (4, 6)
_LAYOUT["Tc"] = (4, 7)
_LAYOUT["Ru"] = (4, 8)
_LAYOUT["Rh"] = (4, 9)
_LAYOUT["Pd"] = (4, 10)
_LAYOUT["Ag"] = (4, 11)
_LAYOUT["Cd"] = (4, 12)
_LAYOUT["In"] = (4, 13)
_LAYOUT["Sn"] = (4, 14)
_LAYOUT["Sb"] = (4, 15)
_LAYOUT["Te"] = (4, 16)
_LAYOUT["I"] = (4, 17)
_LAYOUT["Xe"] = (4, 18)

_LAYOUT["Cs"] = (5, 1)
_LAYOUT["Ba"] = (5, 2)
_LAYOUT["Hf"] = (5, 4)
_LAYOUT["Ta"] = (5, 5)
_LAYOUT["W"] = (5, 6)
_LAYOUT["Re"] = (5, 7)
_LAYOUT["Os"] = (5, 8)
_LAYOUT["Ir"] = (5, 9)
_LAYOUT["Pt"] = (5, 10)
_LAYOUT["Au"] = (5, 11)
_LAYOUT["Hg"] = (5, 12)
_LAYOUT["Tl"] = (5, 13)
_LAYOUT["Pb"] = (5, 14)
_LAYOUT["Bi"] = (5, 15)
_LAYOUT["Po"] = (5, 16)
_LAYOUT["At"] = (5, 17)
_LAYOUT["Rn"] = (5, 18)

_LAYOUT["Fr"] = (6, 1)
_LAYOUT["Ra"] = (6, 2)
_LAYOUT["Rf"] = (6, 4)
_LAYOUT["Db"] = (6, 5)
_LAYOUT["Sg"] = (6, 6)
_LAYOUT["Bh"] = (6, 7)
_LAYOUT["Hs"] = (6, 8)
_LAYOUT["Mt"] = (6, 9)
_LAYOUT["Ds"] = (6, 10)
_LAYOUT["Rg"] = (6, 11)
_LAYOUT["Cn"] = (6, 12)
_LAYOUT["Nh"] = (6, 13)
_LAYOUT["Fl"] = (6, 14)
_LAYOUT["Mc"] = (6, 15)
_LAYOUT["Lv"] = (6, 16)
_LAYOUT["Ts"] = (6, 17)
_LAYOUT["Og"] = (6, 18)

_LAYOUT["La"] = (8, 4)
_LAYOUT["Ce"] = (8, 5)
_LAYOUT["Pr"] = (8, 6)
_LAYOUT["Nd"] = (8, 7)
_LAYOUT["Pm"] = (8, 8)
_LAYOUT["Sm"] = (8, 9)
_LAYOUT["Eu"] = (8, 10)
_LAYOUT["Gd"] = (8, 11)
_LAYOUT["Tb"] = (8, 12)
_LAYOUT["Dy"] = (8, 13)
_LAYOUT["Ho"] = (8, 14)
_LAYOUT["Er"] = (8, 15)
_LAYOUT["Tm"] = (8, 16)
_LAYOUT["Yb"] = (8, 17)
_LAYOUT["Lu"] = (8, 18)

_LAYOUT["Ac"] = (9, 4)
_LAYOUT["Th"] = (9, 5)
_LAYOUT["Pa"] = (9, 6)
_LAYOUT["U"] = (9, 7)
_LAYOUT["Np"] = (9, 8)
_LAYOUT["Pu"] = (9, 9)
_LAYOUT["Am"] = (9, 10)
_LAYOUT["Cm"] = (9, 11)
_LAYOUT["Bk"] = (9, 12)
_LAYOUT["Cf"] = (9, 13)
_LAYOUT["Es"] = (9, 14)
_LAYOUT["Fm"] = (9, 15)
_LAYOUT["Md"] = (9, 16)
_LAYOUT["No"] = (9, 17)
_LAYOUT["Lr"] = (9, 18)

_COLS = range(1, 19)
_ROWS = ["i", "ii", "iii", "iv", "v", "vi", "vii"]

# Dictionary that maps the chemical family with a RGB color.
_FAMILY = {
    "default": (128, 128, 128),
    "user-defined": (255, 255, 255),
    "non metal": (153, 255, 153),
    "noble gas": (192, 255, 255),
    "alkali metal": (255, 153, 153),
    "alkaline earth metal": (255, 222, 173),
    "metalloid": (204, 204, 153),
    "halogen": (255, 255, 153),
    "post-transition metal": (204, 204, 204),
    "transition metal": (255, 192, 192),
    "lanthanoid": (255, 191, 255),
    "actinoid": (255, 153, 254),
}

# Dictionary that maps the chemical state with a RGB color.
_STATE = {
    "default": (128, 128, 128),
    "user-defined": (255, 0, 0),
    "gas": (255, 0, 0),
    "liquid": (0, 0, 255),
    "solid": (0, 0, 0),
    "unknown": (0, 150, 0),
}


class QHLine(QFrame):
    """A placeholder GUI element. Draws a horizontal line."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)


class PopupTextbox(QDialog):
    """A simple dialog displaying text information about a specific
    chemical isotope.
    """

    def __init__(self, *args, input_text="", **kwargs):
        super().__init__(*args, **kwargs)

        self.setStyleSheet("font-family: Courier New;")
        self.text_to_show = input_text
        layout = QVBoxLayout(self)
        self.setLayout(layout)

        # self.mainpart = QTextEdit(self)
        # self.mainpart.append(self.text_to_show)
        temp = QLabel(input_text, self)
        # layout.addWidget(self.mainpart)
        layout.addWidget(temp)

        self.setWindowTitle("Element information summary")
        self.show()
        # self.mainpart.resize(QSize(300,500))


class ElementButton(QToolButton):
    """Subclassed from QToolButton, this object shows the name of a
    chemical element, and creates a pop-up menu giving access to information
    about isotopes when clicked.
    """

    atom_info = Signal(object)
    isotope_info = Signal(str)

    def __init__(self, *args, element="Xx", **kwargs):
        super().__init__(*args, **kwargs)

        self.info = ATOMS_DATABASE[element]
        self.setText(element)
        self.setGroupStyleSheet()
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.isotopes = []

        for iso in ATOMS_DATABASE.get_isotopes(element):
            self.isotopes.append(iso)

        self.clicked.connect(self.altContextMenu)
        self.isotope_info.connect(self.popIstopeInfo)

    @Slot()
    def setGroupStyleSheet(self):
        rgb = _FAMILY[self.info["family"]]
        text_rgb = ",".join([str(int(x)) for x in rgb])
        varbox_stylesheet = (
            "QToolButton {background-color:rgb("
            + text_rgb
            + "); font-size: 20pt ; font-weight: bold; font: ; color: rgb(20,0,0)}"
        )
        self.setStyleSheet(varbox_stylesheet)

    def enterEvent(self, a0: QEnterEvent) -> None:
        self.atom_info.emit(self.info)
        return super().enterEvent(a0)

    def populateMenu(self, menu: QMenu):
        for iso in self.isotopes:
            menu.addAction(iso)

    def altContextMenu(self):
        menu = QMenu()
        self.populateMenu(menu)
        res = menu.exec_(self.mapToGlobal(QPoint(10, 10)))
        if res is not None:
            self.isotope_info.emit(res.text())

    @Slot(str)
    def popIstopeInfo(self, isotope: str):
        infotext = ATOMS_DATABASE.info(isotope)
        PopupTextbox(self, input_text=infotext)

    def contextMenuEvent(self, event):
        menu = QMenu()
        self.populateMenu(menu)
        res = menu.exec_(event.globalPos())
        if res is not None:
            self.isotope_info.emit(res.text())


class InfoDisplay(QFrame):
    """A field embedded in the periodic table of elements.
    Displays information about the chemical elements that the mouse cursor
    is hovering over.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.base = QWidget(self)
        layout = QGridLayout(self.base)
        self.setLayout(layout)

        tooltips = [
            "Atomic number",
            "Symbol",
            "Group,Period,Block",
            "Atom",
            "Chemical family",
            "Electron configuration",
            "Relative atomic mass (uma)",
            "Electronegativity",
            "Electroaffinity (eV)",
            "Ionization energy (eV)",
        ]

        fields = []
        for n in range(10):
            temp = QLabel(self.base)
            fields.append(temp)
            temp.setToolTip(tooltips[n])
        # in the end it was not necessary to have field[0]
        layout.addWidget(fields[1], 0, 0)
        layout.addWidget(fields[2], 1, 0, 1, 4)  # rowspan, columnspan
        layout.addWidget(fields[3], 3, 0)
        for n in range(4, 10):
            layout.addWidget(fields[n], n - 4, 4)

        self.fields = fields

        stylesheet = "font-weight: bold"
        self.setStyleSheet(stylesheet)

    @Slot(object)
    def updateDisplay(self, info_object):
        # self.fields[0].setText(str(info_object["proton"]))
        self.fields[1].setText(
            "<sup>" + str(info_object["proton"]) + "</sup>" + str(info_object["symbol"])
        )
        self.fields[2].setText(
            "%2s,%s,%s"
            % (info_object["group"], info_object["period"], info_object["block"])
        )
        self.fields[3].setText(str(info_object["element"]))
        self.fields[4].setText(str(info_object["family"]))
        self.fields[5].setText(str(info_object["configuration"]))
        self.fields[6].setText(
            "atomic weight = " + str(info_object["atomic_weight"]) + " amu"
        )
        self.fields[7].setText(
            "electronegativity = " + str(info_object["electronegativity"])
        )
        self.fields[8].setText(
            "electroaffinity = " + str(info_object["electroaffinity"]) + " eV"
        )
        self.fields[9].setText(
            "ionization energy = " + str(info_object["ionization_energy"]) + " eV"
        )


class PeriodicTableViewer(QDialog):
    """A widget displaying the periodic table of elements.
    Can be used within MDANSE, or as a standalone program.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("MDANSE Periodic Table of Elements")

        self.glayout = QGridLayout(self)
        self.setLayout(self.glayout)

        self.data_display = InfoDisplay(self)

        self.glayout.addWidget(self.data_display, 1, 5, 3, 8)

        self.placeElements()

    def placeElements(self):
        for key, value in _LAYOUT.items():
            element = ElementButton(self, element=key)
            element.atom_info.connect(self.data_display.updateDisplay)
            row, column = value[0] + 1, value[1] + 1
            self.glayout.addWidget(element, row, column)

        for col_num in range(18):
            self.glayout.addWidget(
                QLabel(str(col_num + 1), self),
                0,
                col_num + 2,
                alignment=Qt.AlignmentFlag.AlignCenter,
            )

        for row_num in range(7):
            self.glayout.addWidget(
                QLabel(_ROWS[row_num], self),
                row_num + 1,
                0,
                alignment=Qt.AlignmentFlag.AlignCenter,
            )

        # now some simple placeholders
        onestar_start = QLabel(self, text="*", alignment=Qt.AlignmentFlag.AlignCenter)
        onestar_end = QLabel(self, text="*", alignment=Qt.AlignmentFlag.AlignCenter)
        twostar_start = QLabel(self, text="**", alignment=Qt.AlignmentFlag.AlignCenter)
        twostar_end = QLabel(self, text="**", alignment=Qt.AlignmentFlag.AlignCenter)
        self.glayout.addWidget(onestar_start, 6, 4)
        self.glayout.addWidget(twostar_start, 7, 4)
        self.glayout.addWidget(onestar_end, 9, 4)
        self.glayout.addWidget(twostar_end, 10, 4)

        self.glayout.addWidget(QHLine(self), 8, 0, 1, 20)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    root = PeriodicTableViewer()
    root.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import typing

from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import QObject, Slot, QMutex
from qtpy.QtWidgets import QTreeView

from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Models.JobTree import JobTree
from MDANSE_GUI.DataViewModel.TrajectoryHolder import DataTreeItem


class ActionsHolder(QStandardItemModel):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.mutex = QMutex()
        self._node_numbers = []
        self._nodes = {}
        self._next_number = 1
        # the nodes used for structuring the data tree
        # will be kept here
        self._sections = {}
        self._section_names = []

    @Slot(object)
    def append_object(self, thing):
        self.mutex.lock()
        self._nodes[self._next_number] = thing
        self._node_numbers.append(self._next_number)
        retval = int(self._next_number)
        self._next_number += 1
        self.mutex.unlock()
        return retval


class ActionsSuperModel(QObject):
    """This object will create and store all the possible
    tree models of actions. These will be assigned to the
    ActionsTree widget, depending on which data object
    has been selected."""

    def __init__(
        self,
        viewer: typing.Optional["QTreeView"] = None,
        parent: typing.Optional["QObject"] = None,
    ) -> None:
        super().__init__(parent)

        self.viewer = viewer
        self.models = {}
        self.currentItem = None
        self.currentModel = None

    def setViewer(self, viewer):
        self.viewer = viewer

    @Slot(DataTreeItem)
    def switchModel(self, item: DataTreeItem):
        """Changes the model behind the viewer to match
        the item selected in the DataView.

        Arguments:
            ancestor -- name of the MDANSE object type
                        selected in the DataView.
        """
        LOG.info(f"received item: {item}")
        ancestor = item.ancestors()[-1]
        try:
            current_model = self.models[ancestor]
        except KeyError:
            current_model = JobTree()
        self.viewer.setModel(current_model)
        self.current_model = current_model
        self.currentItem = item

    def buildModels(self, parent_classes: list):
        """Creates several subtrees out of the registry tree.
        Each tree will only contain nodes that share the same
        ancestor.
        """
        for parent_class in parent_classes:
            model = JobTree()
            model.populateTree(parent_class=parent_class)
            self.models[parent_class.__name__] = model
        LOG.info(f"Build the following models: {self.models.keys()}")
        self.current_model = self.models["IJob"]

    def copyNodeIntoModel(self, thing: typing.Any, model: ActionsHolder):
        node_parents = thing.category
        rootnode = None
        for name in node_parents:
            if name not in model._section_names:
                section_node = QStandardItem()
                section_node.setText(name)
                if rootnode is None:
                    model.appendRow(section_node)
                else:
                    rootnode.appendRow(section_node)
                model._section_names.append(name)
                model._sections[name] = section_node
                rootnode = section_node
            else:
                rootnode = model._sections[name]
        node_text = thing.label
        node_number = model.append_object(thing)
        LOG.info(node_number)
        newitem = QStandardItem()
        newitem.setText(node_text)
        newitem.setData(node_number)
        if rootnode is None:
            model.appendRow(newitem)
        else:
            rootnode.appendRow(newitem)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np

from MDANSE.MolecularDynamics.UnitCell import UnitCell

from MDANSE_GUI.MolecularViewer.readers.i_reader import IReader


class HDF5Wrapper(IReader):
    def __init__(self, fname, trajectory, chemical):
        super(HDF5Wrapper, self).__init__(fname)
        self._n_atoms = chemical._number_of_atoms
        self._n_frames = len(trajectory)
        self._trajectory = trajectory
        self._chemical_system = chemical
        self._filename = fname
        self._atom_ids = [atom.index for atom in chemical.atoms]
        self._atom_types = [atom.symbol for atom in chemical.atoms]
        self._atom_names = [
            "_".join([str(x) for x in [atom.symbol, atom.index]])
            for atom in chemical.atoms
        ]

    def read_frame(self, frame: int) -> "np.array":
        coords = self._trajectory.coordinates(frame)
        return np.array(coords)

    def read_pbc(self, frame: int) -> UnitCell:
        try:
            unit_cell = self._trajectory.unit_cell(frame)
        except AttributeError:
            unit_cell = None
        return unit_cell



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from qtpy.QtWidgets import QSizePolicy, QComboBox, QTableView
from qtpy.QtCore import Slot, Signal, Qt
from qtpy.QtGui import QStandardItemModel, QStandardItem, QBrush

from MDANSE.Framework.QVectors.IQVectors import IQVectors

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class VectorModel(QStandardItemModel):

    type_changed = Signal()
    input_is_valid = Signal(bool)

    def __init__(self, *args, chemical_system=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._generator = None
        self._defaults = []
        self._chemical_system = chemical_system

    @Slot(str)
    def switch_qvector_type(self, vector_type: str, optional_settings: dict = None):
        self.clear()
        self._defaults = []
        self._generator = IQVectors.create(vector_type, self._chemical_system)
        settings = self._generator.settings
        for kv in settings.items():
            name = kv[0]  # dictionary key
            if optional_settings is None:
                value = kv[1][1]["default"]  # tuple value 1: dictionary
            else:
                try:
                    value = optional_settings[name]
                except KeyError:
                    value = kv[1][1]["default"]  # tuple value 1: dictionary
            self._defaults.append(value)
            vtype = kv[1][0]  # tuple value 0: type
            items = [QStandardItem(str(x)) for x in [name, value, vtype]]
            for it in items[0::2]:
                it.setEditable(False)
            for it in items[1::2]:
                it.setData(value, role=Qt.ItemDataRole.ToolTipRole)
            self.appendRow(items)
        self.type_changed.emit()

    def params_summary(self) -> dict:
        params = {}
        all_inputs_are_valid = True
        for rownum in range(self.rowCount()):
            name = str(self.item(rownum, 0).text())
            value = str(self.item(rownum, 1).text())
            vtype = str(self.item(rownum, 2).text())
            try:
                params[name] = self.parse_vtype(vtype, value, name)
            except ValueError:
                params[name] = "failed"
            if params[name] == "failed":
                self.item(rownum, 1).setData(
                    QBrush(Qt.GlobalColor.red), role=Qt.ItemDataRole.BackgroundRole
                )
                all_inputs_are_valid = False
            else:
                self.item(rownum, 1).setData(0, role=Qt.ItemDataRole.BackgroundRole)
        self.input_is_valid.emit(all_inputs_are_valid)
        return params

    def parse_vtype(self, vtype: str, value: str, vname: str):
        if vtype == "RangeConfigurator":
            inner_type = self._generator.settings[vname][1]["valueType"]
            tempstring = value.strip("()[] ")
            result = [inner_type(x) for x in tempstring.split(",")]
            if len(result) == 3:
                return result
            else:
                return "failed"
        elif vtype == "VectorConfigurator":
            inner_type = self._generator.settings[vname][1]["valueType"]
            tempstring = value.strip("()[] ")
            result = [inner_type(x) for x in tempstring.split(",")]
            if len(result) == 3:
                return result
            else:
                return "failed"
        elif vtype == "FloatConfigurator":
            return float(value)
        elif vtype == "IntegerConfigurator":
            return int(value)
        else:
            return value


class QVectorsWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        kwargs["layout_type"] = "QVBoxLayout"
        super().__init__(*args, **kwargs)
        self._relative_size = 3
        trajectory_configurator = kwargs.get("trajectory_configurator", None)
        chemical_system = None
        if trajectory_configurator is not None:
            chemical_system = trajectory_configurator["instance"].chemical_system
        self._selector = QComboBox(self._base)
        self._selector.addItems(IQVectors.indirect_subclasses())
        self._model = VectorModel(self._base, chemical_system=chemical_system)
        self._view = QTableView(self._base)
        self._layout.addWidget(self._selector)
        self._layout.addWidget(self._view)
        self._view.setModel(self._model)
        self._selector.currentTextChanged.connect(self._model.switch_qvector_type)
        self._selector.setCurrentIndex(1)
        self._model.itemChanged.connect(self.updateValue)
        self._model.type_changed.connect(self.updateValue)
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "The parameters needed by the specific q-vector generator can be input here"
        self._view.setToolTip(tooltip_text)
        self._selector.setToolTip(
            "The q vectors will be generated using the method chosen here."
        )
        self._model.input_is_valid.connect(self.validate_model_parameters)
        policy = self._view.sizePolicy()
        policy.setVerticalPolicy(QSizePolicy.Policy.Minimum)
        self._view.setSizePolicy(policy)
        self._view.horizontalHeader().hide()

    @Slot(bool)
    def validate_model_parameters(self, all_are_correct: bool):
        if all_are_correct:
            self.clear_error()
        else:
            self.mark_error("Some entries in the parameter table are invalid.")

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        qvector_type = self._selector.currentText()
        pardict = self._model.params_summary()
        return (qvector_type, pardict)

    def configure_using_default(self):
        """This is too complex to have a default value"""



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import time

from MDANSE.MLogging import LOG

from qtpy.QtCore import QtMsgType
from qtpy.QtCore import qInstallMessageHandler


def generate_timestamp() -> str:
    now = time.gmtime()
    timestamp = (
        "-".join([str(tx) for tx in [now.tm_mday, now.tm_mon, now.tm_year]])
        + ","
        + ":".join([str(ty) for ty in [now.tm_hour, now.tm_min, now.tm_sec]])
    )
    return timestamp


def qt_message_handler(mode, context, message):
    details = ",".join([str(x) for x in [context.line, context.function, context.file]])
    timest = generate_timestamp
    output = ":".join([timest, details, message])
    if mode == QtMsgType.QtInfoMsg:
        LOG.info(output)
    elif mode == QtMsgType.QtWarningMsg:
        LOG.warning(output)
    elif mode == QtMsgType.QtCriticalMsg:
        LOG.critical(output)
    elif mode == QtMsgType.QtFatalMsg:
        LOG.fatal(output)
    else:
        LOG.debug(output)


qInstallMessageHandler(qt_message_handler)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
from collections import defaultdict
from importlib import metadata

from qtpy.QtCore import Slot, QTimer, Signal, QMessageLogger
from qtpy.QtWidgets import (
    QMainWindow,
    QFileDialog,
    QToolBar,
    QTabWidget,
    QMenuBar,
    QMessageBox,
    QApplication,
    QAction,
)

from MDANSE.MLogging import LOG

from MDANSE_GUI.Session.StructuredSession import StructuredSession
from MDANSE_GUI.Resources import Resources
from MDANSE_GUI.UnitsEditor import UnitsEditor
from MDANSE_GUI.UserSettingsEditor import UserSettingsEditor
from MDANSE_GUI.PeriodicTableViewer import PeriodicTableViewer
from MDANSE_GUI.ElementsDatabaseEditor import ElementsDatabaseEditor
from MDANSE_GUI.Tabs.Models.GeneralModel import GeneralModel
from MDANSE_GUI.Tabs.Models.JobHolder import JobHolder
from MDANSE_GUI.Tabs.TrajectoryTab import TrajectoryTab
from MDANSE_GUI.Tabs.JobTab import JobTab
from MDANSE_GUI.Tabs.RunTab import RunTab
from MDANSE_GUI.Tabs.LoggingTab import LoggingTab, GuiLogHandler
from MDANSE_GUI.Tabs.ConverterTab import ConverterTab
from MDANSE_GUI.Tabs.PlotSelectionTab import PlotSelectionTab
from MDANSE_GUI.Tabs.PlotTab import PlotTab
from MDANSE_GUI.Tabs.InstrumentTab import InstrumentTab
from MDANSE_GUI.Widgets.StyleDialog import StyleDialog, StyleDatabase


class TabbedWindow(QMainWindow):
    """The main window of the MDANSE GUI,
    inherits QMainWindow.

    Args:
        QMainWindow - the base class.
    """

    file_name_for_loading = Signal(str)
    converter_name_for_dialog = Signal(str)

    def __init__(
        self,
        *args,
        parent=None,
        title="MDANSE",
        settings=None,
        app_instance=None,
        **kwargs,
    ):
        super().__init__(parent, *args, **kwargs)
        self.tabs = QTabWidget(self)
        self.setCentralWidget(self.tabs)
        self._views = defaultdict(list)
        self._actions = []
        self._tabs = {}
        self._session = StructuredSession()
        self._settings = self._session.obtain_settings(self)
        self._logger = QMessageLogger()
        self._toolbar_buttons = []  # list of (widget, icon_key:str) pairs
        self._style_database = StyleDatabase(self)
        self.setWindowTitle(title)
        self.resources = Resources()
        self.current_object = None
        self.startSettings(settings)
        self.createCommonModels()
        self.makeBasicLayout()
        self.workdir = os.path.expanduser("~")

        self.periodic_table = PeriodicTableViewer(self)
        self.element_editor = ElementsDatabaseEditor(self)
        self.unit_editor = UnitsEditor(self)
        self.style_selector = StyleDialog(self)
        self.style_selector.connectStyleDatabase(self._style_database)
        self.style_selector.new_style.connect(self.setStyleSheet)
        self.style_selector.icon_swap.connect(self.invertToolbar)

        if app_instance is not None:
            app_instance.aboutToQuit.connect(self._session.save)
        self._session.load()
        self.settings_editor = UserSettingsEditor(self, current_session=self._session)

        self._tabs["Plot Creator"]._visualiser.data_for_plotting.connect(
            self._tabs["Plot Holder"].accept_external_data
        )
        self._tabs["Plot Creator"]._visualiser.create_new_plot.connect(
            self._tabs["Plot Holder"]._visualiser.new_plot
        )
        self._tabs["Plot Creator"]._visualiser.create_new_text.connect(
            self._tabs["Plot Holder"]._visualiser.new_text
        )

        self._tabs["Instruments"]._visualiser.instrument_details_changed.connect(
            self._tabs["Actions"].update_action_after_instrument_change
        )

    def createCommonModels(self):
        self._trajectory_model = GeneralModel()
        self._instrument_model = GeneralModel()
        self._job_holder = JobHolder()
        self._gui_log_handler = GuiLogHandler()

    def makeBasicLayout(self):
        self.createConverterViewer()
        self.createTrajectoryViewer()
        self.createActionsViewer()
        self.createJobsViewer()
        self.createPlotSelection()
        self.createPlotHolder()
        self.createInstrumentSelector()
        self.createLogViewer()
        self.setupMenubar()
        self.setupToolbar()

    def startSettings(self, init_settings):
        self.settings = init_settings
        if self.settings is not None:
            self.settings.beginGroup("MainWindow")
            geo = self.settings.value("geometry")
            if geo:
                self.restoreGeometry(geo)
            state = self.settings.value("state")
            if state:
                self.restoreState(state)
            self.settings.endGroup()
        self.settings_timer = QTimer()
        self.settings_timer.timeout.connect(self.saveSettings)
        self.settings_timer.setInterval(2000)
        self.settings_timer.start()
        self.destroyed.connect(self.settings_timer.stop)

    def setupMenubar(self):
        menubar = QMenuBar()
        menubar.setNativeMenuBar(False)  # this works around PyQt problems on MacOS
        menubar.setObjectName("main menubar")
        menubar.setVisible(True)
        file_group = menubar.addMenu("File")
        settings_group = menubar.addMenu("Settings")
        help_group = menubar.addMenu("Help")
        self.exitAct = QAction("Exit", parent=menubar)
        self.exitAct.triggered.connect(self.shut_down)
        file_group.addAction(self.exitAct)
        self.settingsAct = QAction("User Settings", parent=menubar)
        self.settingsAct.triggered.connect(self.launchSettingsEditor)
        settings_group.addAction(self.settingsAct)
        self.aboutAct = QAction("About MDANSE", parent=menubar)
        self.aboutAct.triggered.connect(self.version_information)
        help_group.addAction(self.aboutAct)
        self.setMenuBar(menubar)

    def shut_down(self):
        QApplication.quit()
        self.destroy(True, True)

    def version_information(self):
        version = ""
        version += f"MDANSE version: {metadata.version('MDANSE')}\n"
        version += f"MDANSE_GUI version: {metadata.version('MDANSE_GUI')}\n"
        popup = QMessageBox.about(self, "MDANSE Version Information", version)

    def setupToolbar(self):
        self._toolBar = QToolBar("Main MDANSE toolbar", self)
        # self._toolBar.setMovable(True)
        self._toolBar.setObjectName("main toolbar")
        valid_keys = [
            # ('load', self.loadTrajectory),
            # ('plus', self.loadTrajectory),
            ("periodic_table", self.launchPeriodicTable),
            ("element", self.launchElementsEditor),
            ("units", self.launchUnitsEditor),
            ("user", self.launchStyleSelector),
        ]
        for key, slot in valid_keys:
            icon = self.resources._icons[key]
            action = QAction(icon, str(key), self._toolBar)
            action.triggered.connect(slot)
            self._actions.append(action)
            self._toolbar_buttons.append((action, key))
            # self._actions.append(self._toolBar.addAction(icon, str(key)))
        for act in self._actions:
            self._toolBar.addAction(act)
        self.addToolBar(self._toolBar)
        LOG.info(f"Icon size is {self._toolBar.iconSize()}")

    @Slot()
    def launchPeriodicTable(self):
        self.launch_dialog(self.periodic_table)

    @Slot()
    def launchUnitsEditor(self):
        self.launch_dialog(self.unit_editor)

    @Slot()
    def launchStyleSelector(self):
        self.launch_dialog(self.style_selector)

    @Slot()
    def launchSettingsEditor(self):
        self.settings_editor.update_combo()
        self.launch_dialog(self.settings_editor)

    @Slot()
    def launchElementsEditor(self):
        self.launch_dialog(self.element_editor)

    def launch_dialog(self, dialog) -> None:
        if dialog.isVisible():
            if dialog.isMaximized():
                dialog.showMaximized()
            else:
                dialog.showNormal()
            dialog.activateWindow()
        else:
            dialog.show()

    @Slot()
    def loadTrajectory(self):
        fname = QFileDialog.getOpenFileName(
            self,
            "Load an MD trajectory",
            self.workdir,
            "MDT files (*.mdt);;HDF5 files (*.h5 *.hdf);;All files (*)",
        )
        LOG.info(fname)
        if len(fname[0]) > 0:
            self.file_name_for_loading.emit(fname[0])

    @Slot(bool)
    def invertToolbar(self, dark=False):
        if dark:
            for obj, key in self._toolbar_buttons:
                obj.setIcon(self.resources._inverted_icons[key])
        else:
            for obj, key in self._toolbar_buttons:
                obj.setIcon(self.resources._icons[key])

    def createTrajectoryViewer(self):
        name = "Trajectories"
        trajectory_tab = TrajectoryTab.gui_instance(
            self.tabs,
            name,
            self._session,
            self._settings,
            self._logger,
            model=self._trajectory_model,
        )
        self.tabs.addTab(trajectory_tab._core, name)
        self._tabs[name] = trajectory_tab
        self._job_holder.trajectory_for_loading.connect(trajectory_tab.load_trajectory)

    def createInstrumentSelector(self):
        name = "Instruments"
        instrument_tab = InstrumentTab.gui_instance(
            self.tabs,
            name,
            self._session,
            self._settings,
            self._logger,
            model=self._instrument_model,
        )
        self.tabs.addTab(instrument_tab._core, name)
        self._tabs[name] = instrument_tab

    def createJobsViewer(self):
        name = "Running Jobs"
        run_tab = RunTab.gui_instance(
            self.tabs,
            name,
            self._session,
            self._settings,
            self._logger,
            model=self._job_holder,
        )
        self.tabs.addTab(run_tab._core, name)
        self._tabs[name] = run_tab

    def createConverterViewer(self):
        name = "Converters"
        job_tab = ConverterTab.gui_instance(
            self.tabs, name, self._session, self._settings, self._logger
        )
        job_tab.set_job_starter(self._job_holder)
        self.tabs.addTab(job_tab._core, name)
        self._tabs[name] = job_tab

    def createActionsViewer(self):
        name = "Actions"
        job_tab = JobTab.gui_instance(
            self.tabs,
            name,
            self._session,
            self._settings,
            self._logger,
            combo_model=self._trajectory_model,
            instrument_model=self._instrument_model,
        )
        job_tab.set_job_starter(self._job_holder)
        self.tabs.addTab(job_tab._core, name)
        self._tabs[name] = job_tab
        self.tabs.tabBar().tabBarClicked.connect(
            job_tab.update_action_on_tab_activation
        )

    def createPlotSelection(self):
        name = "Plot Creator"
        plot_tab = PlotSelectionTab.gui_instance(
            self.tabs,
            name,
            self._session,
            self._settings,
            self._logger,
        )
        self.tabs.addTab(plot_tab._core, name)
        self._tabs[name] = plot_tab
        self._job_holder.results_for_loading.connect(plot_tab.load_results)

    def createPlotHolder(self):
        name = "Plot Holder"
        plot_tab = PlotTab.gui_instance(
            self.tabs,
            name,
            self._session,
            self._settings,
            self._logger,
        )
        plot_tab.connect_units()
        self.tabs.addTab(plot_tab._core, name)
        self._tabs[name] = plot_tab

    def createLogViewer(self):
        name = "Logger"
        LOG.addHandler(self._gui_log_handler)
        log_tab = LoggingTab.gui_instance(
            self.tabs, name, self._session, self._settings, self._logger
        )
        log_tab.add_handler(self._gui_log_handler)
        self.tabs.addTab(log_tab._core, name)
        self._tabs[name] = log_tab

    @Slot()
    def saveSettings(self):
        self.settings.beginGroup("MainWindow")
        self.settings.setValue("geometry", self.saveGeometry())
        self.settings.setValue("state", self.saveState())
        self.settings.endGroup()

    def reportError(self, text: str):
        LOG.error(text)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import QAbstractItemModel
from qtpy.QtCore import Slot, Signal, QModelIndex, Qt
from qtpy.QtWidgets import QMenu, QTableView, QAbstractItemView, QMessageBox
from qtpy.QtGui import QStandardItem, QContextMenuEvent

from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Visualisers.TextInfo import TextInfo
from MDANSE_GUI.Tabs.Visualisers.JobLogInfo import JobLogInfo
from MDANSE_GUI.Tabs.Views.Delegates import ProgressDelegate


class RunTable(QTableView):
    item_details = Signal(object)
    jobs_logs = Signal(object)
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.clicked.connect(self.item_picked)
        self._progbar = ProgressDelegate()
        self.setItemDelegateForColumn(1, self._progbar)
        vh = self.verticalHeader()
        vh.setVisible(False)

    def setModel(self, model: QAbstractItemModel) -> None:
        result = super().setModel(model)
        self.model().dataChanged.connect(self.resizeColumnsToContents)
        return result

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        index = self.indexAt(event.pos())
        if index.row() == -1:
            # block right click when it's not on a job
            return
        model = self.model()
        item = model.itemData(index)
        menu = QMenu()
        self.populateMenu(menu, item)
        menu.exec_(event.globalPos())

    def populateMenu(self, menu: QMenu, item: QStandardItem):
        entry, _, _, _ = self.getJobObjects()
        job_state = entry._current_state
        for action, method in [
            ("Delete", self.deleteNode),
            ("Pause", self.pauseJob),
            ("Resume", self.unpauseJob),
            ("Terminate", self.terminateJob),
            # ("Kill", self.killJob),
        ]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)
            if action not in job_state._allowed_actions:
                temp_action.setEnabled(False)

    def getJobObjects(self):
        model = self.model()
        index = self.currentIndex()
        item_row = index.row()
        entry_number = model.index(item_row, 0).data(role=Qt.ItemDataRole.UserRole)
        try:
            entry_number = int(entry_number)
        except ValueError:
            LOG.error(f"Could not use {entry_number} as int")
            return
        job_entry, job_watcher_thread, job_process, log_listener = (
            model.existing_jobs[entry_number],
            model.existing_threads[entry_number],
            model.existing_processes[entry_number],
            model.existing_listeners[entry_number],
        )
        return job_entry, job_watcher_thread, job_process, log_listener

    @Slot()
    def deleteNode(self):
        entry, watcher, process, listener = self.getJobObjects()
        try:
            process.close()
        except ValueError:
            LOG.error("The process is still running!")
        else:
            model = self.model()
            index = self.currentIndex()
            model.removeRow(index.row())
            self.item_details.emit("")
            self.jobs_logs.emit(([], []))

    @Slot()
    def pauseJob(self):
        entry, _, _, _ = self.getJobObjects()
        entry.pause_job()

    @Slot()
    def unpauseJob(self):
        entry, _, _, _ = self.getJobObjects()
        entry.unpause_job()

    @Slot()
    def killJob(self):
        entry, _, process, listener = self.getJobObjects()
        process.kill()
        entry.kill_job()
        listener.stop()

    @Slot()
    def terminateJob(self):
        confirmation_box = QMessageBox(
            QMessageBox.Icon.Question,
            "You are about to terminate a job",
            "All progress will be lost if you terminate your job. Do you really want to teminate?",
            buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            parent=self,
        )
        result = confirmation_box.exec()
        LOG.info(f"QMessageBox result = {result}")
        if result == QMessageBox.StandardButton.Yes:
            entry, _, process, listener = self.getJobObjects()
            # process is not alive, the job probably finished already
            if process.is_alive():
                process.terminate()
                process.join()
                entry.terminate_job()
                listener.stop()

    @Slot(QModelIndex)
    def item_picked(self, index: QModelIndex):
        model = self.model()
        index = self.currentIndex()
        item_row = index.row()
        node_number = model.index(item_row, 0).data(role=Qt.ItemDataRole.UserRole)
        job_entry = model.existing_jobs[node_number]
        self.item_details.emit(job_entry.text_summary())
        self.jobs_logs.emit(job_entry.msgs_and_levels())

    def connect_to_visualiser(self, visualiser: TextInfo) -> None:
        """Connect to a visualiser.

        Parameters
        ----------
        visualiser : TextInfo
            A visualiser to connect to this view.
        """
        if type(visualiser) is TextInfo:
            self.item_details.connect(visualiser.update_panel)
        elif type(visualiser) is JobLogInfo:
            self.jobs_logs.connect(visualiser.update_panel)
        else:
            raise NotImplementedError(
                f"Unable to connect view {type(self)} to visualiser "
                f"{type(visualiser)}"
            )



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QWidget, QVBoxLayout
from qtpy.QtCore import Slot, Signal

from MDANSE_GUI.MolecularViewer.MolecularViewer import MolecularViewer
from MDANSE_GUI.MolecularViewer.Controls import ViewerControls


class View3D(QWidget):
    error = Signal(str)

    def __init__(self, viewer: MolecularViewer, *args, **kwargs):
        super().__init__(*args, **kwargs)

        layout = QVBoxLayout(self)
        self.setLayout(layout)

        controls = ViewerControls(self)
        viewer.setParent(controls)
        controls.setViewer(viewer)
        layout.addWidget(controls)
        self._viewer = viewer
        self._controls = controls

    @Slot(tuple)
    def update_panel(self, data: tuple):
        fullpath, incoming = data
        if fullpath == "" or data is None:
            self._viewer.clear_panel()
            return

        try:
            self._viewer._new_trajectory_object(fullpath, incoming)
        except AttributeError:
            self.error.emit(f"3D View could not visualise {fullpath}")
            self._viewer.clear_trajectory()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List
import csv
import os
from traceback import format_exc

if TYPE_CHECKING:
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext

import numpy as np
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QTextBrowser,
    QGridLayout,
    QLabel,
    QSpinBox,
    QLineEdit,
    QCheckBox,
    QPushButton,
    QComboBox,
    QFileDialog,
)
from qtpy.QtCore import Slot, Signal, Qt

from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Plotters.Plotter import Plotter


class DataWidget(QWidget):

    change_slider_labels = Signal(object)
    change_slider_limits = Signal(object)
    reset_slider_values = Signal(bool)
    change_slider_coupling = Signal(bool)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._plotter = None
        self._sliderpack = None
        self._plotting_context = None
        self._slider_max = 100
        self._current_path = "."
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self.make_toolbar()
        self.make_canvas()
        self.make_bottom_bar()
        self.set_plotter("Text")
        self.update_plotter_params()

    def make_toolbar(self, line_length=6):
        layout = QGridLayout()
        self.layout().addLayout(layout)
        self._preview_widget = QCheckBox(self)
        self._preview_widget.setChecked(True)
        self._preview_widget.checkStateChanged.connect(self.update_plotter_params)
        self._separator_widget = QLineEdit(" ", self)
        self._separator_widget.textChanged.connect(self.update_plotter_params)
        self._comment_widget = QLineEdit("#", self)
        self._comment_widget.textChanged.connect(self.update_plotter_params)
        self._precision_widget = QSpinBox(self)
        self._precision_widget.setValue(5)
        self._precision_widget.valueChanged.connect(self.update_plotter_params)
        self._lines_widget = QSpinBox(self)
        self._lines_widget.setValue(10)
        self._lines_widget.valueChanged.connect(self.update_plotter_params)
        self._columns_widget = QSpinBox(self)
        self._columns_widget.setValue(10)
        self._columns_widget.valueChanged.connect(self.update_plotter_params)
        # populate the layout
        column = 0
        row = 0
        for label_text, widget in [
            ("Preview only (truncate)", self._preview_widget),
            ("Preview lines:", self._lines_widget),
            ("Preview columns:", self._columns_widget),
            ("Rounding precision", self._precision_widget),
            ("Separator character", self._separator_widget),
            ("Comment character", self._comment_widget),
        ]:
            layout.addWidget(QLabel(label_text, self), row, column)
            column += 1
            layout.addWidget(widget, row, column)
            column += 1
            if column >= line_length:
                row += 1
                column = 0

    def make_bottom_bar(self):
        layout = QHBoxLayout()
        self.layout().addLayout(layout)
        layout.addWidget(QLabel("Output file:"))
        self._output_widget = QLineEdit("", self)
        layout.addWidget(self._output_widget)
        self._browse_button = QPushButton("Browse", self)
        self._browse_button.clicked.connect(self.output_file_dialog)
        layout.addWidget(self._browse_button)
        tooltip = (
            "CSV dialect option, as implemented by the Python CSV module. "
            "It should mainly affect delimiters and newline characters. "
            "Most of the time there is no need to change it."
        )
        templabel = QLabel("CSV dialect:", self)
        templabel.setToolTip(tooltip)
        layout.addWidget(templabel)
        self._dialect_combo = QComboBox(self)
        self._dialect_combo.addItems(csv.list_dialects())
        self._dialect_combo.setToolTip(tooltip)
        layout.addWidget(self._dialect_combo)
        self._output_button = QPushButton("Save file", self)
        self._output_button.clicked.connect(self.save_to_file)
        layout.addWidget(self._output_button)

    @Slot()
    def output_file_dialog(self):
        new_value = QFileDialog.getSaveFileName(
            self,  # the parent of the dialog
            "Save data to a CSV file",  # the label of the window
            self._current_path,  # the initial search path
            "Comma-separated values (*.csv);;Output file name (*)",  # text string specifying the file name filter.
        )
        if len(new_value[0]) > 0:
            self._output_widget.setText(new_value[0])
            self._current_path = os.path.split(new_value[0])[0]

    @Slot()
    def save_to_file(self):
        target_path = self._output_widget.text()
        try:
            nsets = len(self._plotter._pc_backup.datasets())
        except AttributeError:
            return
        except:
            LOG.warning("DataWidget could not determine the number of datasets.")
        else:
            if nsets == 0:  # do not create a file if there are no data
                return
        try:
            target = open(target_path, "w", newline="")
        except:
            LOG.error(f"Could not open file for writing: {target_path}")
        else:
            writer = csv.writer(
                target,
                dialect=self._dialect_combo.currentText,
            )
            for header, data in self._plotter._formatter.datasets_for_csv():
                for line in header:
                    target.write(line + "\n")
                for row in data:
                    writer.writerow(row)
            target.close()

    @Slot(object)
    def slider_change(self, new_values: object):
        """Not used for text output."""

    @Slot(bool)
    def set_slider_values(self, reset_needed: bool):
        """Not used for text output."""

    def set_context(self, new_context: "PlottingContext"):
        self._plotting_context = new_context
        self._plotting_context._figure = self._figure

    @Slot(str)
    def set_plotter(self, plotter_option: str):
        try:
            self._plotter = Plotter.create(plotter_option)
        except Exception as e:
            LOG.error(f"DataWidget failed to create Text plotter: {e}")
        else:
            LOG.debug(f"DataWidget created plotter {plotter_option}: {self._plotter}")
            self.plot_data()

    @Slot()
    def update_plotter_params(self):
        if self._plotter is None:
            return
        is_preview = self._preview_widget.isChecked()
        rounding_precision = self._precision_widget.value()
        comment_char = self._comment_widget.text()
        separator_char = self._separator_widget.text()
        preview_lines = self._lines_widget.value()
        preview_columns = self._columns_widget.value()
        self._plotter.adjust_formatter(
            preview=is_preview,
            preview_lines=preview_lines,
            preview_columns=preview_columns,
            rounding=rounding_precision,
            separator=separator_char,
            comment=comment_char,
        )

    @Slot(object)
    def slider_change(self, new_values: object):
        self._plotter.handle_slider(new_values)

    @Slot(bool)
    def set_slider_values(self, reset_needed: bool):
        if reset_needed and self._sliderpack is not None:
            values = self._plotter._initial_values
            self._sliderpack.set_values(values)

    def available_plotters(self) -> List[str]:
        return ["Text"]

    def plot_data(self, update_only=False):
        if self._plotter is None:
            LOG.info("No plotter present in DataWidget.")
            return
        if self._plotting_context is None:
            LOG.info("No plotting context present in DataWidget.")
            return
        try:
            LOG.debug(
                f"DataWidget will plot {self._plotting_context} in {self._figure}."
            )
            self._plotter.plot(
                self._plotting_context,
                self._figure,
                update_only=None,
                toolbar=None,
            )
            for _, databundle in self._plotting_context.datasets().items():
                dataset, _, _, _, _, _ = databundle
                self._current_path = os.path.split(dataset._filename)[0]
                break
        except Exception as e:
            LOG.error(f"DataWidget error: {e}" f"traceback {format_exc()}")

    def make_canvas(self):
        """Creates a matplotlib figure for plotting

        Parameters
        ----------
        width : float, optional
            Figure width in inches, by default 12.0
        height : float, optional
            Figure height in inches, by default 9.0
        dpi : int, optional
            Figure resolution in dots per inch, by default 100

        Returns
        -------
        QWidget
            a widget containing just a QTextBrowser
        """
        layout = self.layout()
        self._figure = QTextBrowser(self)
        layout.addWidget(self._figure)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from MDANSE.Framework.Jobs.IJob import IJob

from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import Signal
from qtpy.QtCore import Qt


class JobTree(QStandardItemModel):
    """Creates a tree structure
    of QStandardItem objects, and stores information
    about the names and docstrings of different
    classes contained in the IJob object.

    It inherits the QStandardItemModel, so it can be
    used in the Qt data/view/proxy model.
    """

    doc_string = Signal(str)
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        parent_class = kwargs.pop("parent_class", IJob)
        filter = kwargs.pop("filter", None)
        super().__init__(*args, **kwargs)

        self._nodes = {}  # dict of {number: QStandardItem}
        self._docstrings = {}  # dict of {number: str}
        self._values = {}  # dict of {number: str}

        self._categories = {}
        self._jobs = {}

        self._by_ancestor = {}  # dict of list[int]

        self.nodecounter = 0  # each node is given a unique number

        self.populateTree(parent_class=parent_class, filter=filter)

    def populateTree(self, parent_class=None, filter=None):
        """This function starts the recursive process of scanning
        the registry tree. Only called once on startup.
        """
        if parent_class is None:
            parent_class = IJob
        full_dict = parent_class.indirect_subclass_dictionary()
        for class_name, class_object in full_dict.items():
            if class_object.enabled:
                self.createNode(class_name, class_object, filter)

    def createNode(self, name: str, thing, filter: str = ""):
        """Creates a new QStandardItem. It will store
        the node number as user data. The 'thing' passed to this method
        will be stored by the model in an internal dictionary, where
        the node number is the key

        Arguments:
            name -- the name of the new node
            thing -- any Python object to be stored and attached to the node
            filter -- a string which must appear in the category tuple
        """
        new_node = QStandardItem(name)
        new_number = self.nodecounter + 1
        self.nodecounter += 1
        new_node.setData(new_number, role=Qt.ItemDataRole.UserRole)
        self._nodes[new_number] = new_node
        self._values[new_number] = thing
        self._docstrings[new_number] = thing.__doc__
        try:
            self._docstrings[new_number] += "\n" + thing.build_doc(use_html_table=True)
        except AttributeError:
            pass
        except TypeError:
            pass
        if hasattr(thing, "category"):
            if filter:
                if filter in thing.category:
                    parent = self.parentsFromCategories(thing.category)
                else:
                    return
            else:
                parent = self.parentsFromCategories(thing.category)
        else:
            parent = self.invisibleRootItem()
        parent.appendRow(new_node)

    def parentsFromCategories(self, category_tuple):
        """Returns the parent node for a node that belongs to the
        category specified by categore_tuple. Also makes sure that
        the parent nodes exist (or creates them if they don't).

        Arguments:
            category_tuple -- category names (str) in the sequence in which
                they should be placed in the tree structure.

        Returns:
            QStandardItem - the node of the last item in 'category_tuple'
        """
        parent = self.invisibleRootItem()
        for cat_string in category_tuple:
            if not cat_string in self._categories.keys():
                current_node = QStandardItem(cat_string)
                parent.appendRow(current_node)
                parent = current_node
                self._categories[cat_string] = current_node
            else:
                current_node = self._categories[cat_string]
                parent = current_node
        return current_node



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QComboBox

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class MultipleCombosWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        configurator = kwargs.get("configurator", None)
        if configurator is None:
            option_list = kwargs.get("choices", [])
            number_of_boxes = 1
        else:
            option_list = configurator.choices
            default_option = configurator.default
            number_of_boxes = configurator.nChoices
        self._fields = []
        for field_number in range(number_of_boxes):
            field = QComboBox(self._base)
            field.addItems(option_list)
            field.setCurrentText(default_option[field_number])
            field.setEditable(False)
            field.currentTextChanged.connect(self.updateValue)
            if self._tooltip:
                tooltip_text = self._tooltip
            else:
                tooltip_text = (
                    "A single option can be picked out of all the options listed."
                )
            field.setToolTip(tooltip_text)
            self._layout.addWidget(field)
            self._fields.append(field)
        # self._field = field
        self._configurator = configurator
        self.default_labels()
        self.update_labels()
        self.updateValue()

    def configure_using_default(self):
        """This is too complex to have a default value"""

    def default_labels(self):
        """Each Widget should have a default tooltip and label,
        which will be set in this method, unless specific
        values are provided in the settings of the job that
        is being configured."""
        if self._label_text == "":
            self._label_text = "ComboWidget"
        if self._tooltip == "":
            self._tooltip = "You only have one option. Choose wisely."

    def get_widget_value(self):
        result = [field.currentText() for field in self._fields]
        return result



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


from qtpy.QtWidgets import (
    QDialog,
    QPushButton,
    QVBoxLayout,
    QApplication,
    QMenu,
    QComboBox,
    QTreeView,
)
from qtpy.QtCore import Signal, Slot, Qt, QSortFilterProxyModel

from MDANSE.MLogging import LOG


class SettingsView(QTreeView):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def populateMenu(self, menu: QMenu, clicked_on_node: bool):
        if clicked_on_node:
            for action, method in [("Delete", self.delete_node)]:
                temp_action = menu.addAction(action)
                temp_action.triggered.connect(method)
            menu.addSeparator()
            for action, method in [("Add new entry", self.append_child)]:
                temp_action = menu.addAction(action)
                temp_action.triggered.connect(method)
        for action, method in [("Add new group", self.append_group)]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)

    def inner_model(self):
        model = self.model()
        try:
            model = model.sourceModel()
        except AttributeError:
            pass
        return model

    @Slot()
    def delete_node(self):
        model = self.inner_model()
        index = self.currentIndex()
        model.removeRow(index.row())

    @Slot()
    def append_child(self):
        model = self.inner_model()
        index = self.currentIndex()
        model.append_child(index.row())

    @Slot()
    def append_group(self):
        model = self.inner_model()
        model.append_group()


class UserSettingsEditor(QDialog):
    def __init__(self, *args, current_session=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("MDANSE User Settings Editor")

        layout = QVBoxLayout(self)

        self.setLayout(layout)

        self.filename_widget = QComboBox(self)
        layout.addWidget(self.filename_widget)
        self._session = current_session

        self.viewer = SettingsView(self)
        self.viewer.setAnimated(True)
        layout.addWidget(self.viewer)

        self.writeout_button = QPushButton("Save settings", self)
        layout.addWidget(self.writeout_button)
        self.writeout_button.clicked.connect(self.save_changes)
        self.filename_widget.currentTextChanged.connect(self.switch_model)
        self.viewer.expanded.connect(self.expand_columns)

    @Slot(str)
    def switch_model(self, model_key: str):
        if not model_key:
            return
        self.data_model = self._session.settings_model(settings_filename=model_key)
        self.data_model.refresh()
        self.proxy_model = QSortFilterProxyModel(self)
        self.proxy_model.setSourceModel(self.data_model)
        self.viewer.setModel(self.proxy_model)
        self.expand_columns()
        self.data_model.itemChanged.connect(self.data_model.on_value_changed)

    @Slot()
    def expand_columns(self):
        for ncol in range(3):
            self.viewer.resizeColumnToContents(ncol)

    def update_combo(self):
        self.filename_widget.clear()
        if self._session is not None:
            self.filename_widget.addItems(
                [str(x) for x in self._session._configs.keys()]
            )
            self.filename_widget.setCurrentText(self._session._main_config_name)

    @Slot()
    def save_changes(self):
        self._session.save()


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    root = UserSettingsEditor()
    root.show()
    app.exec()



import pytest
import tempfile

from MDANSE_GUI.Tabs.Models.JobHolder import JobEntry


@pytest.fixture(scope="module")
def temporary_jobentry() -> JobEntry:
    return JobEntry()


def test_start(temporary_jobentry: JobEntry):
    temporary_jobentry._current_state.start()
    assert temporary_jobentry._current_state._label == "Running"


def test_fail(temporary_jobentry: JobEntry):
    temporary_jobentry._current_state.fail()
    assert temporary_jobentry._current_state._label == "Failed"



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from .DoublePanel import DoublePanel


class MultiPanel(DoublePanel):
    """The multi-panel layout which adds an extra visualiser panels to
    the left or right side of the layout"""

    def __init__(self, *args, **kwargs):
        left_panels = kwargs.pop("left_panels", [])
        right_panels = kwargs.pop("right_panels", [])
        extra_visualiser = kwargs.pop("extra_visualiser", None)
        super().__init__(*args, **kwargs)

        for left_panel in left_panels:
            self._leftlayout.addWidget(left_panel)
            if self._view is not None:
                self._view.connect_to_visualiser(left_panel)

        for right_panel in right_panels:
            self._rightlayout.addWidget(right_panel)
            if self._view is not None:
                self._view.connect_to_visualiser(right_panel)

        self._extra_visualiser = extra_visualiser



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Union

from qtpy.QtCore import Slot, Signal, QModelIndex
from qtpy.QtWidgets import QMenu, QListView, QAbstractItemView
from qtpy.QtGui import QStandardItem, QContextMenuEvent

from MDANSE_GUI.Tabs.Visualisers.View3D import View3D
from MDANSE_GUI.Tabs.Visualisers.TrajectoryInfo import TrajectoryInfo


class TrajectoryView(QListView):
    item_details = Signal(tuple)
    item_name = Signal(str)
    error = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.clicked.connect(self.item_picked)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        index = self.indexAt(event.pos())
        if index.row() == -1:
            # block right click when it's not on a trajectory
            return
        model = self.model()
        item = model.itemData(index)
        menu = QMenu()
        self.populateMenu(menu, item)
        menu.exec_(event.globalPos())

    def populateMenu(self, menu: QMenu, item: QStandardItem):
        for action, method in [("Delete", self.deleteNode)]:
            temp_action = menu.addAction(action)
            temp_action.triggered.connect(method)

    @Slot()
    def deleteNode(self):
        model = self.model()
        index = self.currentIndex()
        model.removeRow(index.row())
        self.item_details.emit(("", None))

    @Slot(QModelIndex)
    def item_picked(self, index: QModelIndex):
        model = self.model()
        node_number = model.itemFromIndex(index).data()
        trajectory = model._nodes[node_number]
        self.item_details.emit(trajectory)

    def connect_to_visualiser(self, visualiser: Union[View3D, TrajectoryInfo]) -> None:
        """Connect to a visualiser.

        Parameters
        ----------
        visualiser : View3D or TrajectoryInfo
            A visualiser to connect to this view.
        """
        if isinstance(visualiser, View3D) or isinstance(visualiser, TrajectoryInfo):
            self.item_details.connect(visualiser.update_panel)
        else:
            raise NotImplementedError(
                f"Unable to connect view {type(self)} to visualiser "
                f"{type(visualiser)}"
            )



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtWidgets import QLineEdit, QLabel

from MDANSE_GUI.InputWidgets.WidgetBase import WidgetBase


class VectorWidget(WidgetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._changing_label = QLabel("Vector", parent=self._base)
        self._layout.addWidget(self._changing_label)
        vfields = []
        for val in ["0", "0", "1"]:
            temp = QLineEdit(val, self._base)
            self._layout.addWidget(temp)
            vfields.append(temp)
        self._vector_fields = vfields
        self._mode = 0
        self.updateValue()
        if self._tooltip:
            tooltip_text = self._tooltip
        else:
            tooltip_text = "The spatial properties in the analysis can be projected on a plane or on an axis, as chosen here."
        for wid in vfields:
            wid.setToolTip(tooltip_text)

    def configure_using_default(self):
        """This is too complex to have a default value"""

    def get_widget_value(self):
        """Collect the results from the input widgets and return the value."""
        try:
            vector = [float(x.text()) for x in self._vector_fields]
        except:
            vector = [0, 0, 0]
        return vector



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List

from matplotlib.pyplot import style as mpl_style
from matplotlib import rcParams

from MDANSE.MLogging import LOG
from MDANSE.Core.SubclassFactory import SubclassFactory
from MDANSE.Framework.Units import measure
from MDANSE_GUI.Tabs.PlotModifiers.NormGenerator import NullNorms
from MDANSE_GUI.Tabs.PlotModifiers.WeightGenerator import NullWeights
from MDANSE_GUI.Tabs.PlotModifiers.AxesWrapper import AxesWrapper

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


class PlotterTemplate(metaclass=SubclassFactory):

    def __init__(self) -> None:
        self._figure = None
        self._axes = {}
        self._unique_axes = []
        self._initial_values = [0.0, 0.0]
        self._slider_values = [0.0, 0.0]
        self._number_of_sliders = 2
        self._value_reset_needed = True
        self._toolbar = None
        self._slider_reference = None
        self._plotting_context = None
        self._weight_generator = NullWeights()
        self._norm_generator = NullNorms()

    def collect_arrays(self):
        (
            self._datasets,
            self._colours,
            self._linestyles,
            self._markers,
            self._labels,
            self._names,
            self._scale_factors,
        ) = ({}, {}, {}, {}, {}, {}, {})
        for name, databundle in self._plotting_context.datasets().items():
            dataset, colour, linestyle, marker, dataset_number, axis_label = databundle
            self._datasets[dataset_number] = dataset
            self._colours[dataset_number] = colour
            self._linestyles[dataset_number] = linestyle
            self._markers[dataset_number] = marker
            self._labels[dataset_number] = axis_label
            self._names[dataset_number] = name
            self._scale_factors[dataset_number] = [1.0]

    def convert_units(self):
        for dataset_number, dataset in self._datasets:
            axis_label = self._labels[dataset_number]
            try:
                best_unit, best_axis = (dataset._axes_units[axis_label], axis_label)
            except KeyError:
                best_unit, best_axis = dataset.longest_axis()
            plotlabel = dataset._labels["medium"]
            xaxis_unit = self._plotting_context.get_conversion_factor(best_unit)
            try:
                conversion_factor = measure(1.0, best_unit, equivalent=True).toval(
                    xaxis_unit
                )
            except:
                LOG.warning(f"Could not convert {best_unit} to {xaxis_unit}")
                conversion_factor = 1.0

    def apply_weights(self):
        for dataset_number, dataset in self._datasets:
            try:
                scale_factor = self._weight_generator(dataset)
            except:
                scale_factor = 1.0
            self._scale_factors[dataset_number].append([scale_factor])

    def apply_norms(self):
        for dataset_number, dataset in self._datasets.items():
            try:
                scale_factor = self._norm_generator(dataset)
            except:
                scale_factor = 1.0
            self._scale_factors[dataset_number].append([scale_factor])

    def reset_figure(self, figure=None, toolbar=None):
        self.enable_slider(False)
        target = self.get_figure(figure)
        if target is None:
            return
        if toolbar is not None:
            self._toolbar = toolbar
        self._figure = target

    def create_plots(self):
        axes = self._figure.add_subplot(111)
        wrapper = AxesWrapper(axes)
        self._unique_axes.append(wrapper)
        for dataset_number, dataset in self._datasets.items():
            self._axes[dataset_number] = wrapper

    def populate_plots(self):
        for dataset_number, dataset in self._datasets.items():
            wrapper = self._axes[dataset_number]
            wrapper.axes.plot(dataset)  # We need to think how to pass information here

    def add_labels(self):
        for wrapper in self._unique_axes:
            wrapper.axes.set_title("MDANSE plot")
            wrapper.axes.grid(True)
            wrapper.axes.legend(loc=0)

    def show_plot(self):
        pass

    def apply_sliders(self):
        slider_values = self._slider_reference._current_values
        self.handle_slider(slider_values)

    def save_settings(self):
        pass

    def plotting_workflow(self):
        self.collect_arrays()
        self.convert_units()
        self.apply_weights()
        self.apply_norms()
        self.reset_figure()
        self.create_plots()
        self.populate_plots()
        self.add_labels()
        self.show_plot()
        self.apply_sliders()
        self.save_settings()

    def request_slider_values(self):
        if self._slider_reference is None:
            return
        self._slider_reference.collect_values()

    def clear(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            return
        target.clear()

    def slider_labels(self) -> List[str]:
        return ["Slider 1", "Slider 2"]

    def slider_limits(self) -> List[str]:
        return self._number_of_sliders * [[-1.0, 1.0, 0.01]]

    def sliders_coupled(self) -> bool:
        return False

    def get_figure(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            LOG.error(f"PlottingContext can't plot to {target}")
            return
        target.clear()
        return target

    def apply_settings(self, plotting_context: "PlottingContext", colours=None):
        if colours is not None:
            self._current_colours = colours
        if plotting_context.set_axes() is None:
            LOG.debug("Axis check failed.")
            return
        try:
            matplotlib_style = colours["style"]
        except:
            pass
        else:
            if matplotlib_style is not None:
                mpl_style.use(matplotlib_style)
        try:
            bkg_col = colours["background"]
        except:
            pass
        else:
            if bkg_col is not None:
                for axes in self._axes:
                    axes.set_facecolor(bkg_col)
        try:
            col_seq = colours["curves"]
        except:
            pass
        else:
            if col_seq is not None:
                for axes in self._axes:
                    axes.set_prop_cycle("color", col_seq)

    def enable_slider(self, allow_slider: bool = True):
        if allow_slider:
            self._slider_reference.setEnabled(True)
            self._slider_reference.blockSignals(False)
        else:
            self._slider_reference.setEnabled(False)
            self._slider_reference.blockSignals(True)

    def handle_slider(self, new_value: List[float]):
        self._slider_values = new_value

    def plot(
        self,
        plotting_context: "PlottingContext",
        figure: "Figure" = None,
        update_only=False,
        toolbar=None,
    ):
        self._plotting_context = plotting_context
        target = self.get_figure(figure)
        if target is None:
            return
        self._target_figure = target
        if toolbar is not None:
            self._toolbar = toolbar
        axes = target.add_subplot(111)
        temp = AxesWrapper(axes)
        self._axes = [temp]
        self.apply_settings(plotting_context)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from qtpy.QtCore import QDir, QSize
from qtpy.QtGui import QIcon, QPixmap

from MDANSE.MLogging import LOG


class Resources:
    def __init__(self):
        self._icons = {}
        self._inverted_icons = {}
        self.load_icons()

    def load_icons(self):
        from importlib.resources import files

        temp = files("MDANSE_GUI")
        # print(f"I got {temp} from the importlib.resources")
        res_dir = QDir(str(temp.joinpath("Icons")))
        LOG.info(f"Resources are in {res_dir.absolutePath()}")
        # res_dir.addSearchPath('icons', 'Src/PyQtGUI/Icons/')
        res_dir.setNameFilters(["*.png"])
        files = res_dir.entryList()
        for f in files:
            label = ".".join(str(f).split(".")[:-1])
            self._icons[label] = QIcon(res_dir.filePath(f))
            # print(f"Loaded {f} from {res_dir}")
        for (
            iname,
            icon,
        ) in self._icons.items():
            pixmap = icon.pixmap(QSize(32, 32))
            image = pixmap.toImage()
            image.invertPixels()
            newpixmap = QPixmap.fromImage(image)
            self._inverted_icons[iname] = QIcon(newpixmap)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List
import math

import numpy as np
from scipy.interpolate import interp1d
from matplotlib.pyplot import colorbar as mpl_colorbar

from MDANSE.Framework.Units import measure
from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Plotters.Plotter import Plotter

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


class Heatmap(Plotter):

    def __init__(self) -> None:
        super().__init__()
        self._figure = None
        self._backup_images = {}
        self._backup_arrays = {}
        self._backup_minmax = {}
        self._backup_scale_interpolators = {}
        self._backup_limits = {}
        self._last_axes_units = {}
        self._initial_values = [0.0, 100.0]
        self._slider_values = [0.0, 100.0]
        self._last_minmax = [-1, -1]
        self._slice_axis = 2
        self._plot_limit = 1

    def clear(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            return
        target.clear()

    def slider_labels(self) -> List[str]:
        return ["Minimum (percentile)", "Maximum (percentile)"]

    def slider_limits(self) -> List[str]:
        return self._number_of_sliders * [[0.0, 100.0, 0.1]]

    def sliders_coupled(self) -> bool:
        return True

    def get_figure(self, figure: "Figure" = None):
        if figure is None:
            target = self._figure
        else:
            target = figure
        if target is None:
            LOG.error(f"PlottingContext can't plot to {target}")
            return
        target.clear()
        return target

    def handle_slider(self, new_value: List[float]):
        super().handle_slider(new_value)
        target = self._figure
        if target is None:
            return
        if new_value[1] <= new_value[0]:
            return
        else:
            self._slider_values = [new_value[0], new_value[1]]
        for ds_num, image in self._backup_images.items():
            try:
                last_minmax = self._backup_minmax[ds_num]
            except KeyError:
                self._backup_minmax[ds_num] = [-1, -1]
                last_minmax = [-1, -1]
            interpolator = self._backup_scale_interpolators[ds_num]
            newmax = interpolator(new_value[1])
            newmin = interpolator(new_value[0])
            if newmax < newmin:
                if newmax == last_minmax[1]:
                    newmin = float(newmax)
                else:
                    newmax = float(newmin)
            if newmin == last_minmax[0] and newmax == last_minmax[1]:
                return
            if newmax >= newmin:
                try:
                    image.set_clim([newmin, newmax])
                except ValueError:
                    LOG.error(
                        f"Matplotlib could not set colorbar limits to {newmin}, {newmax}"
                    )
                else:
                    self._figure.canvas.draw_idle()
                    self._backup_minmax[ds_num] = [newmin, newmax]
        target.canvas.draw()

    def plot(
        self,
        plotting_context: "PlottingContext",
        figure: "Figure" = None,
        update_only=False,
        toolbar=None,
    ):
        self.enable_slider(True)
        target = self.get_figure(figure)
        if target is None:
            return
        if toolbar is not None:
            self._toolbar = toolbar
        self._figure = target
        self._backup_images = {}
        self._backup_arrays = {}
        self._backup_scale_interpolators = {}
        xaxis_unit = None
        yaxis_unit = None
        self._axes = []
        if not update_only:
            self._last_axes_units = {}
        self.apply_settings(plotting_context)
        if plotting_context.set_axes() is None:
            LOG.debug("Axis check failed.")
            return
        nplots = 0
        for databundle in plotting_context.datasets().values():
            ds, _, _, _, ds_num, axis_label = databundle
            if ds._n_dim == 1:
                continue
            elif ds._n_dim == 3:
                replacement_axis_number = None
                for number, axis_name in enumerate(ds._axes.keys()):
                    if axis_name == axis_label:
                        replacement_axis_number = number
                if replacement_axis_number is None:
                    ds.planes_vs_axis(self._slice_axis)
                else:
                    ds.planes_vs_axis(replacement_axis_number)
                nplots += len(ds._planes)
            else:
                nplots += 1
            try:
                self._backup_scale_interpolators[ds_num](51.2)
            except:
                percentiles = np.linspace(0, 100.0, 21)
                results = [np.percentile(ds._data, perc) for perc in percentiles]
                self._backup_scale_interpolators[ds_num] = interp1d(
                    percentiles, results
                )
        if nplots > self._plot_limit:
            nplots = self._plot_limit
        gridsize = int(math.ceil(nplots**0.5))
        startnum = 1
        for num, databundle in enumerate(plotting_context.datasets().values()):
            dataset, _, _, _, ds_num, _ = databundle
            transposed = False
            if dataset._n_dim == 1:
                continue
            if dataset._n_dim == 3:
                all_numbers, all_datasets = (
                    list(dataset._planes.keys()),
                    list(dataset._planes.values()),
                )
                all_labels = [dataset._plane_labels[number] for number in all_numbers]
            else:
                primary_axis_number = 0
                for number, axis_name in enumerate(ds._axes.keys()):
                    if axis_name == axis_label:
                        primary_axis_number = number
                all_numbers = [0]
                if primary_axis_number == 0:
                    all_datasets = [dataset._data.T]
                else:
                    all_datasets = [dataset._data]
                    transposed = True
                all_labels = [dataset._name]
            limits = []
            axis_units = []
            for key, value in dataset._axes_units.items():
                target_unit = plotting_context.get_conversion_factor(value)
                axis_array = dataset._axes[key]
                try:
                    conversion_factor = measure(1.0, value, equivalent=True).toval(
                        target_unit
                    )
                except:
                    limits += [axis_array[0], axis_array[-1]]
                    axis_units.append(value)
                else:
                    limits += [
                        axis_array[0] * conversion_factor,
                        axis_array[-1] * conversion_factor,
                    ]
                    axis_units.append(target_unit)
            if transposed:
                axis_units = axis_units[::-1]
                limits = limits[2:] + limits[:2]
            for xnum in range(len(all_datasets)):
                if startnum > self._plot_limit:
                    break
                axes = target.add_subplot(gridsize, gridsize, startnum)
                startnum += 1
                self._axes.append(axes)
                image = axes.imshow(
                    all_datasets[xnum][::-1, :],
                    extent=limits,
                    aspect="auto",
                    interpolation=None,
                    cmap=plotting_context.colormap,
                )
                axes.set_title(all_labels[xnum])
                colorbar = mpl_colorbar(image, ax=image.axes, format="%.1e", pad=0.02)
                colorbar.set_label(dataset._data_unit)
                xlimits, ylimits = axes.get_xlim(), axes.get_ylim()
            self._backup_arrays[ds_num] = np.nan_to_num(dataset._data)
            if update_only:
                interpolator = self._backup_scale_interpolators[ds_num]
                last_minmax = [
                    interpolator(self._slider_values[0]),
                    interpolator(self._slider_values[1]),
                ]
                try:
                    image.set_clim(last_minmax)
                except ValueError:
                    LOG.error(
                        f"Matplotlib could not set colorbar limits to {last_minmax}"
                    )
                try:
                    last_limits = self._backup_limits[ds_num]
                except KeyError:
                    self._backup_limits[ds_num] = [
                        xlimits[0],
                        xlimits[1],
                        ylimits[0],
                        ylimits[1],
                    ]
                    last_limits = self._backup_limits[ds_num]
                else:
                    try:
                        previous_units = self._last_axes_units[ds_num]
                    except KeyError:
                        self._last_axes_units[ds_num] = [axis_units[0], axis_units[1]]
                        previous_units = self._last_axes_units[ds_num]
                    if axis_units[0] == previous_units[0]:
                        try:
                            axes.set_xlim((last_limits[0], last_limits[1]))
                        except ValueError:
                            LOG.error(
                                f"Matplotlib could not set x limits to {last_limits[0]}, {last_limits[1]}"
                            )
                    else:
                        self._last_axes_units[ds_num][0] = axis_units[0]
                        xlim = axes.get_xlim()
                        self._backup_limits[ds_num][0] = xlim[0]
                        self._backup_limits[ds_num][1] = xlim[1]
                    if axis_units[1] == previous_units[1]:
                        try:
                            axes.set_ylim((last_limits[2], last_limits[3]))
                        except ValueError:
                            LOG.error(
                                f"Matplotlib could not set x limits to {last_limits[2]}, {last_limits[3]}"
                            )
                    else:
                        self._last_axes_units[ds_num][1] = axis_units[1]
                        ylim = axes.get_ylim()
                        self._backup_limits[ds_num][2] = ylim[0]
                        self._backup_limits[ds_num][3] = ylim[1]
            else:
                try:
                    last_limits = self._backup_limits[ds_num]
                except KeyError:
                    self._backup_limits[ds_num] = [
                        xlimits[0],
                        xlimits[1],
                        ylimits[0],
                        ylimits[1],
                    ]
                    last_limits = self._backup_limits[ds_num]
                interpolator = self._backup_scale_interpolators[ds_num]
                last_minmax = [
                    interpolator(self._slider_values[0]),
                    interpolator(self._slider_values[1]),
                ]
                try:
                    image.set_clim(last_minmax)
                except ValueError:
                    LOG.error(
                        f"Matplotlib could not set colorbar limits to {last_minmax}"
                    )
                self._backup_minmax[ds_num] = [dataset._data.min(), dataset._data.max()]
                self._backup_limits[ds_num] = [
                    xlimits[0],
                    xlimits[1],
                    ylimits[0],
                    ylimits[1],
                ]
                self._last_axes_units[ds_num] = axis_units
            axes.set_xlabel(axis_units[0])
            axes.set_ylabel(axis_units[1])
            self._backup_images[ds_num] = image
        # axes.grid(True)
        self.request_slider_values()
        target.canvas.draw()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import os
from functools import partial

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QWidget

from MDANSE import PLATFORM
from MDANSE.MLogging import LOG
from MDANSE.Framework.InputData.HDFTrajectoryInputData import HDFTrajectoryInputData

from MDANSE_GUI.Tabs.GeneralTab import GeneralTab
from MDANSE_GUI.Tabs.Layouts.DoublePanel import DoublePanel
from MDANSE_GUI.Session.LocalSession import LocalSession
from MDANSE_GUI.Tabs.Models.GeneralModel import GeneralModel
from MDANSE_GUI.Tabs.Views.InstrumentList import InstrumentList
from MDANSE_GUI.Tabs.Visualisers.InstrumentDetails import InstrumentDetails


label_text = """Here you can browse, edit and add instrument profiles.
If you are trying to reproduce the results of a neutron experiment,
you will need to set at least the correct instrument resolution
and q-vector coverage to be used in the analysis.
The initial inputs of an analysis in the GUI will be affected
by the instrument profile you chose. You can still change them
before starting the analysis, if you had something else in mind.
"""


class InstrumentTab(GeneralTab):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._core.add_button("Create Instrument", self._view.add_instrument)
        self._core.add_button("Save Instruments", self._view.save_to_file, upper=False)
        self._view.introduce_empty_instrument()
        current_path = os.path.dirname(os.path.abspath(__file__))
        builtin_file = os.path.join(
            current_path, "..", "Resources", "InstrumentDefinitions.toml"
        )
        try:
            self._view.load_from_file(builtin_file, keep_backups=True)
        except Exception as e:
            LOG.error(f"Could not load instruments from {builtin_file}: {e}")
        filename = os.path.join(
            PLATFORM.application_directory(), "InstrumentDefinitions.toml"
        )
        try:
            self._view.load_from_file(filename)
        except Exception as e:
            LOG.error(f"Could not load instruments from {filename}: {e}")
        for instrument in self._model._nodes.values():
            if instrument is not None:
                instrument.update_item()
                instrument._configured = (
                    True  # instruments loaded from file are configured
                )

    @Slot()
    def load_trajectories(self):
        self._view.add_instrument()

    @Slot(str)
    def load_trajectory(self, fname: str):
        if len(fname) > 0:
            _, short_name = os.path.split(fname)
            try:
                data = HDFTrajectoryInputData(fname)
            except Exception as e:
                self._core.error.emit(repr(e))
            else:
                self._core._model.append_object(((fname, data), short_name))

    @classmethod
    def standard_instance(cls):
        the_tab = cls(
            window,
            name="Instruments",
            session=LocalSession(),
            model=GeneralModel(),
            view=InstrumentList(),
            visualiser=InstrumentDetails(),
            layout=DoublePanel,
            label_text=label_text,
        )
        return the_tab

    @classmethod
    def gui_instance(
        cls,
        parent: QWidget,
        name: str,
        session: LocalSession,
        settings,
        logger,
        **kwargs,
    ):
        the_tab = cls(
            parent,
            name=name,
            session=session,
            settings=settings,
            logger=logger,
            model=kwargs.get("model", GeneralModel()),
            view=InstrumentList(),
            visualiser=InstrumentDetails(),
            layout=DoublePanel,
            label_text=label_text,
        )
        return the_tab


if __name__ == "__main__":
    import sys
    from qtpy.QtWidgets import QApplication, QMainWindow

    app = QApplication(sys.argv)
    window = QMainWindow()
    the_tab = InstrumentTab.standard_instance()
    window.setCentralWidget(the_tab._core)
    window.show()
    app.exec()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, List
import math

from MDANSE.Framework.Units import measure
from MDANSE.MLogging import LOG

from MDANSE_GUI.Tabs.Plotters.Plotter import Plotter

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from MDANSE_GUI.Tabs.Models.PlottingContext import PlottingContext


class Grid(Plotter):

    def __init__(self) -> None:
        super().__init__()
        self._figure = None
        self._backup_limits = []
        self._plot_limit = 8

    def slider_labels(self) -> List[str]:
        return ["Inactive", "Inactive"]

    def slider_limits(self) -> List[str]:
        return self._number_of_sliders * [[-1.0, 1.0, 0.01]]

    def plot(
        self,
        plotting_context: "PlottingContext",
        figure: "Figure" = None,
        update_only=False,
        toolbar=None,
    ):
        self.enable_slider(False)
        target = self.get_figure(figure)
        if target is None:
            return
        if toolbar is not None:
            self._toolbar = toolbar
        if plotting_context.set_axes() is None:
            LOG.debug("Axis check failed.")
            return
        self._axes = []
        self.apply_settings(plotting_context)
        nplots = 0
        for databundle in plotting_context.datasets().values():
            ds, colour, linestyle, marker, _, axis_label = databundle
            try:
                best_unit, best_axis = ds._axes_units[axis_label], axis_label
            except KeyError:
                best_unit, best_axis = ds.longest_axis()
            curves = ds.curves_vs_axis(best_unit)
            nplots += len(curves)
        if nplots > self._plot_limit:
            nplots = self._plot_limit
        gridsize = int(math.ceil(nplots**0.5))
        startnum = 1
        counter = 0
        for name, databundle in plotting_context.datasets().items():
            if counter > self._plot_limit:
                break
            dataset, colour, linestyle, marker, ds_num, axis_label = databundle
            try:
                best_unit, best_axis = dataset._axes_units[axis_label], axis_label
            except KeyError:
                best_unit, best_axis = dataset.longest_axis()
            xaxis_unit = plotting_context.get_conversion_factor(best_unit)
            for key, curve in dataset._curves.items():
                if counter > self._plot_limit:
                    break
                counter += 1
                axes = target.add_subplot(gridsize, gridsize, startnum)
                self._axes.append(axes)
                plotlabel = dataset._labels["medium"]
                conversion_factor = measure(1.0, best_unit, equivalent=True).toval(
                    xaxis_unit
                )
                [temp_curve] = axes.plot(
                    dataset._axes[best_axis] * conversion_factor,
                    curve,
                    linestyle=linestyle,
                    color=colour,
                    label=plotlabel + ":" + dataset._curve_labels[key],
                )
                try:
                    temp_curve.set_marker(marker)
                except ValueError:
                    try:
                        temp_curve.set_marker(int(marker))
                    except:
                        pass
                xlimits, ylimits = axes.get_xlim(), axes.get_ylim()
                try:
                    new_limits = self._backup_limits[ds_num]
                except IndexError:
                    while len(self._backup_limits) < (ds_num + 1):
                        self._backup_limits.append(
                            [
                                xlimits[0],
                                xlimits[1],
                                ylimits[0],
                                ylimits[1],
                            ]
                        )
                    new_limits = self._backup_limits[ds_num]
                if not update_only:
                    self._backup_limits[ds_num] = [
                        xlimits[0],
                        xlimits[1],
                        ylimits[0],
                        ylimits[1],
                    ]
                else:
                    try:
                        axes.set_xlim((new_limits[0], new_limits[1]))
                    except ValueError:
                        LOG.error(
                            f"Matplotlib could not set x limits to {new_limits[0]}, {new_limits[1]}"
                        )
                    try:
                        axes.set_ylim((new_limits[2], new_limits[3]))
                    except ValueError:
                        LOG.error(
                            f"Matplotlib could not set y limits to {new_limits[2]}, {new_limits[3]}"
                        )
                axes.grid(True)
                axes.set_xlabel(xaxis_unit)
                axes.legend(loc=0)
                startnum += 1
        self.apply_settings(plotting_context)
        target.canvas.draw()



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from logging import Handler
from logging.handlers import QueueListener
from multiprocessing import Pipe, Queue, Event
import traceback

from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import QObject, Slot, Signal, QTimer, QThread, QMutex, Qt

from MDANSE.MLogging import FMT, LOG
from MDANSE.Framework.Converters import Converter

from MDANSE_GUI.Subprocess.Subprocess import Subprocess, Connection
from MDANSE_GUI.Subprocess.JobState import (
    Starting,
    Finished,
    Running,
    Failed,
    Paused,
    Aborted,
)
from MDANSE_GUI.Subprocess.JobStatusProcess import JobCommunicator
from MDANSE_GUI.Tabs.Views.Delegates import ProgressDelegate


class JobThread(QThread):

    def __init__(
        self,
        job_comm: "JobCommunicator",
        receiving_end: "Connection",
        subprocess_reference: "Subprocess",
    ):
        super().__init__()
        self._job_comm = job_comm
        self._pipe_end = receiving_end
        self._subprocess = subprocess_reference
        self._keep_running = True
        self._timer = QTimer()
        self._timer.timeout.connect(self.check_if_alive)
        self._timer.setInterval(2000)

    def start(self, *args, **kwargs) -> None:
        retval = super().start(*args, **kwargs)
        self._timer.start()
        return retval

    def fail(self):
        self._job_comm.status_update(("COMMUNICATION", False))
        self._keep_running = False
        self._timer.stop()
        self.terminate()

    @Slot()
    def check_if_alive(self):
        if self._subprocess._closed:
            # The subprocess was closed probably by the user, don't need
            # to keep checking that the subprocess is alive anymore.
            # Also, no need to send out a status update since it should
            # already have been updated already when it got terminated.
            self._keep_running = False
            self._timer.stop()
            self.terminate()
            return
        if not self._subprocess.is_alive():
            self.fail()

    def run(self):
        while self._keep_running:
            try:
                status_update = self._pipe_end.recv()
            except:
                self.fail()
            else:
                self._job_comm.status_update(status_update)


class JobEntry(Handler, QObject):
    """This coordinates all the objects that make up one line on the list
    of current jobs. It is used for reporting the task progress to the GUI."""

    for_loading = Signal(str)

    def __init__(
        self,
        *args,
        command=None,
        entry_number=0,
        pause_event=None,
        load_afterwards=False,
    ):
        super().__init__()
        QObject.__init__(self)
        self._command = command
        self._parameters = {}
        self._pause_event = pause_event
        self._load_afterwards = load_afterwards
        # state pattern
        self._current_state = Starting(self)
        self._Starting = Starting(self)
        self._Finished = Finished(self)
        self._Aborted = Aborted(self)
        self._Running = Running(self)
        self._Failed = Failed(self)
        self._Paused = Paused(self)
        # other variables
        self.percent_complete = 0
        self.steps_complete = 0
        self._entry_number = entry_number
        self.total_steps = 99
        self._prog_item = QStandardItem()
        self._stat_item = QStandardItem()
        for item in [self._stat_item]:
            item.setData(entry_number)
        self._prog_item.setData(0, role=Qt.ItemDataRole.UserRole)
        self._prog_item.setData("progress", role=Qt.ItemDataRole.DisplayRole)
        self._prog_item.setData(0, role=ProgressDelegate.progress_role)
        self._prog_item.setData(100, role=ProgressDelegate.progress_role + 1)
        self.records = []

    def text_summary(self) -> str:
        result = ""
        result += f"Job type: {self._command}\n"
        result += "Parameters:\n"
        for key, value in self._parameters.items():
            result += f" - {key} = {value}\n"
        result += "Status:\n"
        result += f"Current state: {self._current_state._label}\n"
        result += f"Percent complete: {self.percent_complete}\n"
        return result

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, input: dict):
        self._parameters = input

    def update_fields(self):
        self._prog_item.setText(f"{self.percent_complete} percent complete")
        self._prog_item.setData(self.percent_complete, role=Qt.ItemDataRole.UserRole)
        self._prog_item.setData(
            int(self.steps_complete), role=ProgressDelegate.progress_role
        )
        self._stat_item.setText(self._current_state._label)

    @Slot(bool)
    def on_finished(self, success: bool):
        if success:
            self._current_state.finish()
            if self._load_afterwards:
                try:
                    len(self._parameters["output_files"][1])
                except TypeError:  # job is a converter
                    file_name = self._parameters["output_files"][0]
                    if ".mdt" not in file_name[-5:]:
                        file_name += ".mdt"
                    self.for_loading.emit(file_name)
                else:  # job is an analysis
                    if "MDAFormat" in self._parameters["output_files"][1]:
                        file_name = self._parameters["output_files"][0]
                        if ".mda" not in file_name[-5:]:
                            file_name += ".mda"
                        self.for_loading.emit(file_name)
        else:
            self._current_state.fail()
        self.update_fields()

    @Slot(int)
    def on_started(self, target_steps: int):
        LOG.info(f"Item received on_started: {target_steps} total steps")
        self.total_steps = target_steps
        self._prog_item.setData(target_steps, role=ProgressDelegate.progress_role + 1)
        self._current_state.start()
        self.update_fields()

    @Slot(int)
    def on_update(self, completed_steps: int):
        # print(f"completed {completed_steps} out of {self.total_steps} steps")
        if self.total_steps > 0:
            self.steps_complete = completed_steps
            self.percent_complete = round(99 * completed_steps / self.total_steps, 1)
        else:
            self.percent_complete = 0
        self.update_fields()
        self._prog_item.emitDataChanged()

    @Slot()
    def on_oscillate(self):
        """For jobs with unknown duration, the progress bar will bounce."""

    def pause_job(self):
        self._current_state.pause()
        self.update_fields()

    def unpause_job(self):
        self._current_state.unpause()
        self.update_fields()

    def terminate_job(self):
        self._current_state.terminate()
        self.update_fields()

    def kill_job(self):
        self._current_state.kill()
        self.update_fields()

    def msgs_and_levels(self):
        msgs = []
        levels = []
        for record in self.records:
            msgs.append(self.format(record))
            levels.append(record.levelname)
        return msgs, levels

    def emit(self, record):
        self.records.append(record)


class JobHolder(QStandardItemModel):
    """All the job INSTANCES that are started by the GUI
    are added to this model."""

    trajectory_for_loading = Signal(str)
    results_for_loading = Signal(str)

    def __init__(self, parent: QObject = None):
        super().__init__(parent=parent)
        self.lock = QMutex()
        self.existing_threads = {}
        self.existing_processes = {}
        self.existing_jobs = {}
        self.existing_listeners = {}
        self._next_number = 0
        self.setHorizontalHeaderLabels(["Job", "Progress", "Status"])

    @Slot(str)
    def reportError(self, err: str):
        LOG.error(err)

    @property
    def next_number(self):
        retval = int(self._next_number)
        self._next_number += 1
        return retval

    @Slot(list)
    def startProcess(self, job_vars: list, load_afterwards=False):
        log_queue = Queue()

        main_pipe, child_pipe = Pipe()
        pause_event = Event()
        entry_number = self.next_number

        item_th = JobEntry(
            command=job_vars[0],
            entry_number=entry_number,
            pause_event=pause_event,
            load_afterwards=load_afterwards,
        )
        item_th.setFormatter(FMT)
        item_th.setLevel("INFO")

        try:
            subprocess_ref = Subprocess(
                job_name=job_vars[0],
                job_parameters=job_vars[1],
                pipe=child_pipe,
                pause_event=pause_event,
                log_queue=log_queue,
            )
        except Exception as e:
            LOG.error(
                f"Failed to create Subprocess using {job_vars};\n"
                f"error {e};\n"
                f"traceback {traceback.format_exc()}"
            )
            return

        handlers = [item_th] + LOG.handlers
        listener = QueueListener(log_queue, *handlers, respect_handler_level=True)
        listener.start()

        communicator = JobCommunicator()
        watcher_thread = JobThread(communicator, main_pipe, subprocess_ref)
        communicator.moveToThread(watcher_thread)
        entry_number = self.next_number
        item_th.parameters = job_vars[1]
        if load_afterwards:
            if job_vars[0] in Converter.subclasses():
                item_th.for_loading.connect(self.trajectory_for_loading)
            else:
                try:
                    int(job_vars[1]["output_files"][1])
                except:
                    item_th.for_loading.connect(self.results_for_loading)
                else:
                    item_th.for_loading.connect(self.trajectory_for_loading)
        communicator.target.connect(item_th.on_started)  # int
        communicator.progress.connect(item_th.on_update)  # int
        communicator.finished.connect(item_th.on_finished)  # bool
        communicator.oscillate.connect(item_th.on_oscillate)  # nothing
        LOG.info("Watcher thread ready to start!")
        watcher_thread.start()
        try:
            task_name = str(job_vars[0])
        except:
            task_name = str("This should have been a job name")
        name_item = QStandardItem(task_name)
        name_item.setData(entry_number, role=Qt.ItemDataRole.UserRole)
        self.appendRow(
            [
                name_item,
                item_th._prog_item,
                item_th._stat_item,
            ]
        )
        # nrows = self.rowCount()
        # index = self.indexFromItem(item_th._item)
        # print(f"Index: {index}")
        self.existing_processes[entry_number] = subprocess_ref
        self.existing_threads[entry_number] = watcher_thread
        self.existing_jobs[entry_number] = item_th
        self.existing_listeners[entry_number] = listener
        LOG.info("Subprocess ready to start!")
        subprocess_ref.start()

    def startProcessAndLoad(self, job_vars: list):
        self.startProcess(job_vars, load_afterwards=True)



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
import abc
import collections

import numpy as np

from MDANSE.MLogging import LOG
from MDANSE.Chemistry import ATOMS_DATABASE
from MDANSE.Chemistry import RESIDUES_DATABASE


class InvalidFileError(Exception):
    """This class implements an exception for invalid file."""


class InvalidAtomError(Exception):
    """This class implements an exception raised in case of an invalid atom selection."""


class IReader(abc.ABC):
    """This class implements an interface for trajectory readers."""

    def __init__(self, filename):
        """Constructor.

        Args:
            filename (str): the trajectory filename
        """

        self._filename = filename

        self._n_frames = 0

        self._n_atoms = 0

    @property
    def filename(self):
        return self._filename

    @property
    def n_atoms(self):
        return self._n_atoms

    @property
    def atom_ids(self):
        return self._atom_ids

    @property
    def atom_names(self):
        return self._atom_names

    @property
    def atom_types(self):
        return self._atom_types

    @property
    def molecules(self):
        mol_indexes = collections.OrderedDict()

        for i, resid in enumerate(self._residue_ids):
            mol_indexes.setdefault(resid, []).append(i)

        return list(mol_indexes.values())

    @property
    def residue_ids(self):
        return self._residue_ids

    @property
    def residue_names(self):
        return self._residue_names

    @property
    def n_frames(self):
        return self._n_frames

    @property
    def times(self):
        return self._times

    @abc.abstractmethod
    def read_frame(self, frame):
        pass

    @abc.abstractmethod
    def read_pbc(self, frame):
        pass

    def get_atom_indexes(self, residue_names, atom_names):
        """Return the nested list of the indexes of the atoms whose residue and name are respectively in the provided
        list of residue and atom names.

        Args:
            residue_names (list of str): the residues to scan
            atom_names (list of str): the atoms to scan
        """

        indexes = []
        for i, at in enumerate(self._atom_names):
            current_residue = self._residue_names[i]
            if current_residue not in residue_names:
                continue

            if at in atom_names:
                indexes.append(i)

        indexes_per_molecule = collections.OrderedDict()
        for idx in indexes:
            resid = self._residue_ids[idx]
            indexes_per_molecule.setdefault(resid, []).append(idx)

        indexes_per_molecule = list(indexes_per_molecule.values())

        return indexes_per_molecule

    def guess_atom_types(self):
        """Guess the atom type (element) from their atom names.

        For standard residues, the strategy is to start from the left and search
        by increasing length until a valid element is found.
        For unknown residue, the strategy is opposite. Indeed, we start from the
        right and search by decreasing length until a valid element is found.
        The elemenents are searched in an internal YAML database.
        """

        # Retrieve all the chemical symbols from the internal database
        symbols = [at["symbol"].upper() for at in ATOMS_DATABASE]

        self._atom_types = []
        for i in range(self._n_atoms):
            atom_name = self._atom_names[i]
            residue_name = self._residue_names[i]

            # Remove the trailing and initial digits from the upperized atom names
            upper_atom_name = atom_name.upper()
            upper_atom_name = upper_atom_name.lstrip("0123456789").rstrip("0123456789")

            # Case of the an atom that belongs to a standard residue
            # Guess the atom type by the starting from the first alpha letter from the left,
            # increasing the word by one letter if there was no success in guessing the atom type
            if residue_name in RESIDUES_DATABASE:
                start = 1
                while True:
                    upper_atom_name = upper_atom_name[:start]
                    if upper_atom_name in symbols:
                        self._atom_types.append(upper_atom_name.capitalize())
                        break
                    if start > len(atom_name):
                        raise ValueError("Unknown atom type: {}".format(atom_name))
                    start += 1
            # Case of the an atom that does not belong to a standard residue
            # Guess the atom type by the starting from whole atom name,
            # decreasing the word by one letter from the right if there was no success in guessing the atom type
            else:
                start = len(upper_atom_name)
                while True:
                    upper_atom_name = upper_atom_name[:start]
                    if upper_atom_name in symbols:
                        self._atom_types.append(upper_atom_name.capitalize())
                        break
                    if start == 0:
                        raise ValueError("Unknown atom type: {}".format(atom_name))
                    start -= 1

    def read_atom_trajectory(self, index):
        """Read the trajectory of a single atom with a given index

        Returns:
            3-tuple: the coordinates of the atoms through the trajectory alongside with the lower bounds and upper bounds
        """

        if index < 0 or index >= self._n_atoms:
            raise InvalidAtomError("Invalid atom index")

        coords = []
        lower_bounds = []
        upper_bounds = []
        for f in range(self._n_frames):
            frame = self.read_frame(f)
            coords.append(frame[index, :])
            lower_bounds.append(frame.min(axis=0))
            upper_bounds.append(frame.max(axis=0))

        coords = np.array(coords)
        lower_bounds = np.array(lower_bounds)
        upper_bounds = np.array(upper_bounds)

        return coords, lower_bounds, upper_bounds



#    This file is part of MDANSE_GUI.
#
#    MDANSE_GUI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import TYPE_CHECKING, Dict, List
import os
import itertools

if TYPE_CHECKING:
    import h5py

import numpy as np
from matplotlib.markers import MarkerStyle
from matplotlib.lines import lineStyles
from matplotlib import rcParams
from matplotlib.colors import to_hex as mpl_to_hex
import matplotlib.pyplot as mpl
from qtpy.QtCore import Slot, Signal, QModelIndex, Qt
from qtpy.QtGui import QStandardItemModel, QStandardItem, QColor

from MDANSE.MLogging import LOG


def get_mpl_markers():
    unique_keys = list(set(MarkerStyle.markers.keys()))
    filtered_markers = {
        key: MarkerStyle.markers[key]
        for key in unique_keys
        if key not in ["", " ", "none"] + [str(x) for x in range(10)]
    }
    return filtered_markers


def get_mpl_lines():
    filtered_line_styles = {
        key: value for key, value in lineStyles.items() if key not in ["", " "]
    }
    return filtered_line_styles


def get_mpl_colours():
    cycler = rcParams["axes.prop_cycle"]
    colours = cycler.by_key()["color"]
    return [mpl_to_hex(col) for col in colours]


class SingleDataset:

    def __init__(self, name: str, source: "h5py.File"):
        self._name = name
        self._filename = source.filename
        self._curves = {}
        self._curve_labels = {}
        self._planes = {}
        self._plane_labels = {}
        self._data_limits = None
        bare_name = os.path.split(self._filename)[-1]
        self._labels = {
            "minimal": name,
            "medium": f"{bare_name}:{name}",
            "full": f"{self._filename}:{name}",
        }
        self._valid = True
        try:
            self._data = source[name][:]
        except KeyError:
            LOG.debug(f"{name} is not a data set in the file")
            self._valid = False
            return
        except TypeError:
            self._valid = False
            LOG.debug(f"{name} is not plottable")
            return
        self._data_unit = source[name].attrs["units"]
        self._n_dim = len(self._data.shape)
        self._axes_tag = source[name].attrs["axis"]
        self._axes = {}
        self._axes_units = {}
        for ax_number, axis_name in enumerate(self._axes_tag.split("|")):
            aname = axis_name.strip()
            if aname == "index":
                self._axes[aname + str(ax_number)] = np.arange(len(self._data))
                self._axes_units[aname + str(ax_number)] = "N/A"
            else:
                self._axes[aname] = source[aname][:]
                self._axes_units[aname] = source[aname].attrs["units"]

    def set_data_limits(self, limit_string: str):
        complete_subset_list = []
        for token in limit_string.split(";"):
            if ":" in token:
                try:
                    slice_parts = [int(x) for x in token.split(":")]
                except:
                    continue
                if len(slice_parts) < 4:
                    complete_subset_list += list(range(*slice_parts))
            elif "-" in token:
                try:
                    slice_parts = [int(x) for x in token.split("-")]
                except:
                    continue
                if len(slice_parts) == 2:
                    complete_subset_list += list(range(slice_parts[0], slice_parts[1]))
            elif "," in token:
                try:
                    slice_parts = [int(x) for x in token.split(",")]
                except:
                    continue
                complete_subset_list += list(slice_parts)
            else:
                try:
                    complete_subset_list += [int(token)]
                except:
                    continue
        if len(complete_subset_list) == 0:
            self._data_limits = None
        else:
            self._data_limits = np.unique(complete_subset_list).astype(int)

    def available_x_axes(self) -> List[str]:
        return list(self._axes_units.keys())

    def longest_axis(self) -> str:
        length = -1
        best_unit = "none"
        best_axis = "none"
        for aname, aunit in self._axes_units.items():
            xaxis = self._axes[aname]
            xlen = len(xaxis)
            if xlen > length:
                length = xlen
                best_unit = aunit
                best_axis = aname
        return best_unit, best_axis

    def curves_vs_axis(self, axis_unit: str) -> List[np.ndarray]:
        self._curves = {}
        self._curve_labels = {}
        found = -1
        total_ndim = len(self._data.shape)
        data_shape = self._data.shape
        for aname, aunit in self._axes_units.items():
            if aunit == axis_unit:
                xaxis = self._axes[aname]
                xlen = len(xaxis)
                for dim in range(total_ndim):
                    if xlen == self._data.shape[dim]:
                        found = dim
                        break
        slicer = []
        indexer = []
        for dim in range(total_ndim):
            if dim == found:
                slicer.append([slice(None)])
                # indexer.append([None])
            else:
                slicer.append(np.arange(data_shape[dim]))
                indexer.append(np.arange(data_shape[dim]))
        indices = list(itertools.product(*indexer))
        slicers = list(itertools.product(*slicer))
        for n in range(len(indices)):
            if self._data_limits is not None:
                if n in self._data_limits:
                    self._curves[tuple(indices[n])] = self._data[slicers[n]].squeeze()
                    self._curve_labels[tuple(indices[n])] = str(tuple(indices[n]))
            else:
                self._curves[tuple(indices[n])] = self._data[slicers[n]].squeeze()
                self._curve_labels[tuple(indices[n])] = str(tuple(indices[n]))
        return self._curves
        # slicer = tuple(slicer)
        # temp = self._data[slicer].squeeze()
        # for line in temp:
        #     if len(line) != xlen:
        #         print("Wrong data length in the curves_vs_axis method of PlottingContext")
        # return temp

    def planes_vs_axis(self, axis_number: int) -> List[np.ndarray]:
        self._planes = {}
        self._plane_labels = {}
        found = -1
        total_ndim = len(self._data.shape)
        if total_ndim == 1:
            return
        elif total_ndim == 2:
            return self._data
        data_shape = self._data.shape
        number_of_planes = data_shape[axis_number]
        perpendicular_axis = None
        perpendicular_axis_name = ""
        slice_def = []
        for number, (axis_name, axis_array) in enumerate(self._axes.items()):
            if number == axis_number:
                slice_def.append(0)
                perpendicular_axis = axis_array
                perpendicular_axis_name = axis_name
            else:
                slice_def.append(slice(None))
        for plane_number in range(number_of_planes):
            if self._data_limits is not None:
                if plane_number in self._data_limits:
                    fixed_argument = perpendicular_axis[plane_number]
                    slice_def[axis_number] = plane_number
                    data = self._data[tuple(slice_def)]
                    self._planes[plane_number] = data
                    self._plane_labels[plane_number] = (
                        f"{perpendicular_axis_name}={fixed_argument}"
                    )
            else:
                fixed_argument = perpendicular_axis[plane_number]
                slice_def[axis_number] = plane_number
                data = self._data[tuple(slice_def)]
                self._planes[plane_number] = data
                self._plane_labels[plane_number] = (
                    f"{perpendicular_axis_name}={fixed_argument}"
                )


class SingleCurve:

    def __init__(self, data_name: str, file_name: str, *args, **kwargs):
        self._name = data_name
        self._filename = file_name
        self._curve = []
        self._x_axis = []
        self._y_axis = None
        self._z_axis = None
        bare_name = os.path.split(self._filename)[-1]
        self._labels = {
            "minimal": data_name,
            "medium": f"{bare_name}:{data_name}",
            "full": f"{file_name}:{data_name}",
        }
        self._label = self._labels["medium"]
        self._units = {"data": "arbitrary", "x": None, "y": None, "z": None}
        self._colour = "#000000"
        self._line = "-"

    def set_data(self, array: np.ndarray, unit: str):
        self._curve = array
        self._units["data"] = unit

    def set_x_axis(self, array: np.ndarray, unit: str):
        self._x_axis = array
        self._units["x"] = unit

    def set_y_axis(self, value: float, unit: str):
        self._y_axis = value
        self._units["y"] = unit

    def set_z_axis(self, value: float, unit: str):
        self._z_axis = value
        self._units["z"] = unit

    def standard_items(self, key: int) -> List["QStandardItem"]:
        result = []
        for val in []:
            item = QStandardItem(str(val))
            item.setData(key, role=Qt.ItemDataRole.UserRole)
            result.append(item)
        return result


plotting_column_labels = [
    "Dataset",
    "Trajectory",
    "Size",
    "Unit",
    "Main axis",
    "Use it?",
    "Colour",
    "Line style",
    "Marker",
]
plotting_column_index = {
    label: number for number, label in enumerate(plotting_column_labels)
}


class PlottingContext(QStandardItemModel):

    needs_an_update = Signal()

    def __init__(self, *args, unit_lookup=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._datasets = {}
        self._current_axis = [None, None, None]
        self._figure = None
        self._ndim_lowest = 1
        self._ndim_highest = 3
        self._all_xunits = []
        self._best_xunits = []
        self._colour_list = get_mpl_colours()
        self._last_colour_list = get_mpl_colours()
        self._colour_map = kwargs.get("colormap", "viridis")
        self._last_colour = 0
        self._unit_lookup = unit_lookup
        self.setHorizontalHeaderLabels(plotting_column_labels)

    def generate_colour(self, number: int):
        return self._colour_list[number % len(self._colour_list)]

    def next_colour(self):
        colour = self.generate_colour(self._last_colour)
        self._last_colour += 1
        return colour

    @property
    def colormap(self):
        backup_cmap = "viridis"
        try:
            cmap = self._unit_lookup._settings.group("colours").get("colormap")
        except:
            return backup_cmap
        else:
            if cmap in mpl.colormaps():
                return cmap
            else:
                return self._unit_lookup._settings.default_value("colours", "colormap")

    @Slot()
    def regenerate_colours(self):
        self._colour_list = get_mpl_colours()
        self._last_colour = 0
        for row in range(self.rowCount()):
            current_colour = self.item(row, plotting_column_index["Colour"]).text()
            if (
                current_colour
                != self._last_colour_list[row % len(self._last_colour_list)]
            ):
                LOG.debug(
                    f"colours not equal: {current_colour} vs {self._last_colour_list[row % len(self._last_colour_list)]}"
                )
                self._last_colour += 1
            else:
                next_colour = self.next_colour()
                self.item(row, plotting_column_index["Colour"]).setText(
                    str(next_colour)
                )
                self.item(row, plotting_column_index["Colour"]).setData(
                    QColor(str(next_colour)), role=Qt.ItemDataRole.BackgroundRole
                )
        self._last_colour_list = list(self._colour_list)

    @Slot(object)
    def accept_external_data(self, other: "PlottingContext"):
        """Crucial slot for transferring data
        between tabs. DataPlotter will send the datasets
        into this slot."""
        for dataset in other._datasets.values():
            self.add_dataset(dataset)
        self.set_axes()

    @Slot(dict)
    def accept_units(self, units: Dict):
        """Crucial slot for transferring data
        between tabs. DataPlotter will send the datasets
        into this slot."""
        self._unit_preference = units
        self.set_axes()

    def get_conversion_factor(self, unit):
        factor, new_unit = self._unit_lookup.conversion_factor(unit)
        return new_unit

    def datasets(self) -> Dict:
        result = {}
        for ds_num, row in enumerate(range(self.rowCount())):
            key = self.index(row, plotting_column_index["Dataset"]).data(
                role=Qt.ItemDataRole.UserRole
            )
            useit = (
                self.itemFromIndex(
                    self.index(row, plotting_column_index["Use it?"])
                ).checkState()
                == Qt.CheckState.Checked
            )
            data_number_string = self.itemFromIndex(
                self.index(row, plotting_column_index["Use it?"])
            ).text()
            colour = self.itemFromIndex(
                self.index(row, plotting_column_index["Colour"])
            ).text()
            style = self.itemFromIndex(
                self.index(row, plotting_column_index["Line style"])
            ).text()
            marker = self.itemFromIndex(
                self.index(row, plotting_column_index["Marker"])
            ).text()
            axis = self.itemFromIndex(
                self.index(row, plotting_column_index["Main axis"])
            ).text()
            if useit:
                self._datasets[key].set_data_limits(data_number_string)
                result[key] = (self._datasets[key], colour, style, marker, ds_num, axis)
            else:
                self._datasets[key]._data_limits = None
        return result

    def add_dataset(self, new_dataset: SingleDataset):
        if not new_dataset._valid:
            return
        newkey = f"{new_dataset._filename}:{new_dataset._name}"
        if newkey in self._datasets:
            return
        self._datasets[newkey] = new_dataset
        items = [
            QStandardItem(str(x))
            for x in [
                new_dataset._name,
                new_dataset._filename,
                new_dataset._data.shape,
                new_dataset._data_unit,
                new_dataset.longest_axis()[-1],
                "",
                self.next_colour(),
                "-",
                "",
            ]
        ]
        for item in items:
            item.setData(newkey, role=Qt.ItemDataRole.UserRole)
        for item in items[:4]:
            item.setEditable(False)
        temp = items[plotting_column_index["Use it?"]]
        temp.setCheckable(True)
        temp.setCheckState(Qt.CheckState.Checked)
        self.itemChanged.connect(self.needs_an_update)
        temp = items[plotting_column_index["Colour"]]
        temp.setData(QColor(temp.text()), role=Qt.ItemDataRole.BackgroundRole)
        # test for possible nested items
        best_axis = new_dataset.longest_axis()
        curves = new_dataset.curves_vs_axis(best_axis[0])
        if len(curves) > 1:
            counter = 1
            for indices, curve in curves.items():
                temp = SingleCurve(new_dataset._name, new_dataset._filename)
                temp.set_data(curve, new_dataset._data_unit)
                temp.set_x_axis(new_dataset._axes[best_axis[1]], best_axis[0])
                # optionally, we should set y and z as well
                subitems = temp.standard_items(newkey + f":{counter}")
                items[0].appendRow(subitems)
                counter += 1
        elif len(curves) == 1:
            LOG.debug("A single curve output from PlottingContext.")
        else:
            LOG.debug("No curves!")
        self.appendRow(items)

    def set_axes(self):
        if len(self._datasets) == 0:
            return
        all_axes = []
        all_dimensions = []
        longest_axes = []
        unit_to_axname = {}
        for dataset in self._datasets.values():
            n_dimensions = dataset._n_dim
            if n_dimensions not in all_dimensions:
                all_dimensions.append(n_dimensions)
            best_units, best_axis = dataset.longest_axis()
            unit_to_axname[best_units] = best_axis
            for unit in dataset._axes_units.values():
                if unit not in all_axes:
                    all_axes.append(unit)
            if best_units not in longest_axes:
                longest_axes.append(best_units)
        self._ndim = 1
        self._ndim_highest = np.max(all_dimensions)
        self._ndim_lowest = np.min(all_dimensions)
        self._best_xunits = longest_axes
        self._all_xunits = all_axes
        self._unit_to_axname = unit_to_axname
        # self.needs_an_update.emit()
        return "Configured!"

    @Slot()
    def clear(self) -> None:
        result = super().clear()
        self._datasets = {}
        self.setHorizontalHeaderLabels(plotting_column_labels)
        return result

    @Slot(QModelIndex)
    def delete_dataset(self, index: QModelIndex):
        dkey = index.data(role=Qt.ItemDataRole.UserRole)
        self.removeRow(index.row())
        self._datasets.pop(dkey, None)
